/*
 *    MIRACL  C++ Header file elliptic.h
 *
 *    AUTHOR  : M. Scott
 *  
 *    PURPOSE : Definition of class ECn  (Arithmetic on an Elliptic Curve,
 *               mod n)
 *
 *    NOTE    : Must be used in conjunction with elliptic.cpp and big.cpp
 *              The active curve is set dynamically (via the Big ecurve() 
 *              routine) - so beware the pitfalls implicit in declaring
 *              static or global ECn's (which are initialised before the 
 *              curve is set!). Uninitialised data is OK 
 *
 *    Copyright (c) 1988-1997 Shamus Software Ltd.
 */

#ifndef ELLIPTIC_H
#define ELLIPTIC_H

#include <big.h>

class ECn
{
    epoint *p;
public:
    ECn()                         { p=epoint_init();}
   
    ECn(Big &x,Big& y)              {p=epoint_init(); 
                                   epoint_set(x.getbig(),y.getbig(),0,p); }
    
  // This next constructor restores a point on the curve from "compressed" 
  // data, that is the full x co-ordinate, and the LSB of y  (0 or 1)

    ECn(Big& x,int cb)             {p=epoint_init();
                                   epoint_set(x.getbig(),x.getbig(),cb,p); }
                   
    ECn(ECn &b)                   {p=epoint_init(); epoint_copy(b.p,p);}

    
    ECn& operator=(const ECn& b)  {epoint_copy(b.p,p);return *this;}

    ECn& operator+=(const ECn& b) {ecurve_add(b.p,p); return *this;}

  // Multiplication of a point by an integer. 

    ECn& operator*=(Big& k) {ecurve_mult(k.getbig(),p,p); return *this;}

    BOOL set(Big& x,Big& y)    {return epoint_set(x.getbig(),y.getbig(),0,p);}
    int get(Big& x,Big& y)     {return epoint_get(p,x.getbig(),y.getbig());}

  // point compression

  // This sets the point from compressed form. cb is LSB of y coordinate 

    BOOL set(Big& x,int cb=0)  {return epoint_set(x.getbig(),x.getbig(),cb,p);}
  // This gets the point in compressed form. Return value is LSB of y-coordinate

    int get(Big& x)            {return epoint_get(p,x.getbig(),x.getbig());}
  
    friend ECn mul(Big& ,const ECn& ,Big& ,const ECn& );
    friend ECn mul(int, Big *, ECn *);
  
    friend void normalise(ECn &e) {epoint_norm(e.p);}

    friend BOOL operator==(const ECn& a,const ECn& b)
                                  {return epoint_comp(a.p,b.p);}    
    friend BOOL operator!=(const ECn& a,const ECn& b)
                                  {return (!epoint_comp(a.p,b.p));}    

    friend ECn operator*(Big &,const ECn&);
    friend ostream& operator<<(ostream& ,ECn& );

    ~ECn() {epoint_free(p); }

};

#endif

