/*
 * Implementation of the NIST Advanced Ecryption Standard
 *
 * Could/Should be speeded up by assembly patches in time critical loops in
 * aes_encrypt() and aes_decrypt()
 *
 * Note that the AES winner has not yet been determined! Nonetheless
 * we are jumping the gun a little and implementing one of the favourites
 * RIJNDAEL, partly because it is entirely patent-free.
 *
 *
 * If another competitor should emerge as the AES winner, it will be 
 * implemented here instead.
 */

#include <stdio.h>
#include <miracl.h>

#define BYTE unsigned char
#define WORD mr_unsign32

/* this is fixed */
#define NB 4

/* define this for a quick space/time performance boost */
/* #define LOTS_OF_MEMORY    */

/* rotates x one bit to the left */

#define ROTL(x) (((x)>>7)|((x)<<1))

/* Rotates 32-bit word left by 1, 2 or 3 byte  */

#define ROTL8(x) (((x)<<8)|((x)>>24))
#define ROTL16(x) (((x)<<16)|((x)>>16))
#define ROTL24(x) (((x)<<24)|((x)>>8))

static WORD ftable[256];
static WORD rtable[256];
static WORD rco[16];

#ifdef LOTS_OF_MEMORY  

static WORD ftable1[256];
static WORD ftable2[256];
static WORD ftable3[256];

static WORD rtable1[256];
static WORD rtable2[256];
static WORD rtable3[256];

#endif

static BYTE InCo[4]={0xB,0xD,0x9,0xE};  /* Inverse Coefficients */

static BYTE fbsub[256];
static BYTE rbsub[256];
static BYTE ptab[256],ltab[256];


static BOOL TABLES=FALSE;

static WORD pack(BYTE *b)
{ /* pack bytes into a 32-bit Word */
    return (b[3]<<24)|(b[2]<<16)|(b[1]<<8)|b[0];
}

static void unpack(WORD a,BYTE *b)
{ /* unpack bytes from a word */
    b[0]=(BYTE)a;
    b[1]=(BYTE)(a>>8);
    b[2]=(BYTE)(a>>16);
    b[3]=(BYTE)(a>>24);
}

static BYTE xtime(BYTE a)
{
    BYTE b;
    if (a&0x80) b=0x1B;
    else        b=0;
    a<<=1;
    a^=b;
    return a;
}

static BYTE bmul(BYTE x,BYTE y)
{ /* x.y= AntiLog(Log(x) + Log(y)) */
    if (x && y) return ptab[(ltab[x]+ltab[y])%255];
    else return 0;
}

static WORD SubByte(WORD a)
{
    BYTE b[4];
    unpack(a,b);
    b[0]=fbsub[b[0]];
    b[1]=fbsub[b[1]];
    b[2]=fbsub[b[2]];
    b[3]=fbsub[b[3]];
    return pack(b);    
}

static BYTE product(WORD x,WORD y)
{ /* dot product of two 4-byte arrays */
    BYTE xb[4],yb[4];
    unpack(x,xb);
    unpack(y,yb); 
    return bmul(xb[0],yb[0])^bmul(xb[1],yb[1])^bmul(xb[2],yb[2])^bmul(xb[3],yb[3]);
}

static WORD InvMixCol(WORD x)
{ /* matrix Multiplication */
    WORD y,m;
    BYTE b[4];

    m=pack(InCo);
    b[3]=product(m,x);
    m=ROTL24(m);
    b[2]=product(m,x);
    m=ROTL24(m);
    b[1]=product(m,x);
    m=ROTL24(m);
    b[0]=product(m,x);
    y=pack(b);
    return y;
}

static void gentables(void)
{ /* generate tables */
    int i;
    BYTE x,y,b[4];

  /* use 3 as primitive root to generate power and log tables */

    ltab[0]=0;
    ptab[0]=1;  ltab[1]=0;
    ptab[1]=3;  ltab[3]=1; 
    for (i=2;i<256;i++)
    {
        ptab[i]=ptab[i-1]^xtime(ptab[i-1]);
        ltab[ptab[i]]=i;
    }
    
  /* affine transformation:- each bit is xored with itself shifted one bit */

    fbsub[0]=0x63;
    rbsub[0x63]=0;
    for (i=1;i<256;i++)
    {
        y=ptab[255-ltab[i]];  /* multiplicative inverse */
        x=y;  x=ROTL(x);
        y^=x; x=ROTL(x);
        y^=x; x=ROTL(x);
        y^=x; x=ROTL(x);
        y^=x; y^=0x63;
        fbsub[i]=y; rbsub[y]=i;
    }

    for (i=0,y=1;i<16;i++)
    {
        rco[i]=y;
        y=xtime(y);
    }

  /* calculate forward and reverse tables */
    for (i=0;i<256;i++)
    {
        y=fbsub[i];
        b[3]=y^xtime(y); b[2]=y;
        b[1]=y;          b[0]=xtime(y);
        ftable[i]=pack(b);

        y=rbsub[i];
        b[3]=bmul(InCo[0],y); b[2]=bmul(InCo[1],y);
        b[1]=bmul(InCo[2],y); b[0]=bmul(InCo[3],y);
        rtable[i]=pack(b);

#ifdef LOTS_OF_MEMORY
        ftable1[i]=ROTL8(ftable[i]);
        ftable2[i]=ROTL16(ftable[i]);
        ftable3[i]=ROTL24(ftable[i]);

        rtable1[i]=ROTL8(rtable[i]);
        rtable2[i]=ROTL16(rtable[i]);
        rtable3[i]=ROTL24(rtable[i]);
#endif

    }
}

void aes_reset(aes *a,char *iv)
{ /* reset mode, or reset iv */
    int i,j;
    if (iv==NULL) 
    {
        for (i=0;i<NB;i++) a->chain[i]=0;
        a->CBC=FALSE;
        return;
    }
    else
    {
        a->CBC=TRUE;
        for (i=j=0;i<NB;i++,j+=4)
            a->chain[i]=pack(&iv[j]);
    }
}

BOOL aes_init(aes* a,int nk,char *key,char *iv)
{ /* Key=nk bytes */
  /* currently NB,nk = 16, 24 or 32          */
  /* Key Scheduler. Create expanded encryption key */
    int i,j,k,m,n,N,nr,C1,C2,C3;
    WORD CipherKey[8];
    
    nk/=4;

    if (nk!=4 && nk!=6 && nk!=8) return FALSE;

  /* nr is number of rounds */
    nr=6+nk;

    a->fkey=(WORD *)mr_alloc(NB*(nr+1),4);
    if (a->fkey==NULL) return FALSE;
    a->rkey=(WORD *)mr_alloc(NB*(nr+1),4);
    if (a->rkey==NULL) 
    {
        mr_free(a->fkey);
        return FALSE;
    }

    if (!TABLES)
    { /* generate fixed tables just once ever */
        gentables();
        TABLES=TRUE;
    }
    
    a->Nk=nk; a->Nr=nr;

    aes_reset(a,iv);

    N=NB*(nr+1);
    
    for (i=j=0;i<nk;i++,j+=4)
    {
        CipherKey[i]=pack(&key[j]);
    }
    for (i=0;i<nk;i++) a->fkey[i]=CipherKey[i];
    for (j=nk,k=0;j<N;j+=nk,k++)
    {
        a->fkey[j]=a->fkey[j-nk]^SubByte(ROTL24(a->fkey[j-1]))^rco[k];
        if (nk<=6)
        {
            for (i=1;i<nk && (i+j)<N;i++)
                a->fkey[i+j]=a->fkey[i+j-nk]^a->fkey[i+j-1];
        }
        else
        {
            for (i=1;i<4 && (i+j)<N;i++)
                a->fkey[i+j]=a->fkey[i+j-nk]^a->fkey[i+j-1];
            if ((j+4)<N) a->fkey[j+4]=a->fkey[j+4-nk]^SubByte(a->fkey[j+3]);
            for (i=5;i<nk && (i+j)<N;i++)
                a->fkey[i+j]=a->fkey[i+j-nk]^a->fkey[i+j-1];
        }
    }

 /* now for the expanded decrypt key in reverse order */

    for (j=0;j<NB;j++) a->rkey[j+N-NB]=a->fkey[j]; 
    for (i=NB;i<N-NB;i+=NB)
    {
        k=N-NB-i;
        for (j=0;j<NB;j++) a->rkey[k+j]=InvMixCol(a->fkey[i+j]);
    }
    for (j=N-NB;j<N;j++) a->rkey[j-N+NB]=a->fkey[j];

    return TRUE;
}

void aes_encrypt(aes* a,char *buff)
{
    int i,j,k; 
    WORD p[4],q[4],*x,*y,*t;

    for (i=j=0;i<NB;i++,j+=4)
    {
        p[i]=pack(&buff[j]);
        if (a->CBC) p[i]^=a->chain[i];
        p[i]^=a->fkey[i];
    }

    k=NB;
    x=p; y=q;

/* State alternates between p and q */
    for (i=1;i<a->Nr;i++)
    { /* Nr is number of rounds. May be odd. */
#ifdef LOTS_OF_MEMORY
        y[0]=a->fkey[k]^ftable[(BYTE)x[0]]^
             ftable1[(BYTE)(x[1]>>8)]^
             ftable2[(BYTE)(x[2]>>16)]^
             ftable3[x[3]>>24];
        y[1]=a->fkey[k+1]^ftable[(BYTE)x[1]]^
             ftable1[(BYTE)(x[2]>>8)]^
             ftable2[(BYTE)(x[3]>>16)]^
             ftable3[x[0]>>24];
        y[2]=a->fkey[k+2]^ftable[(BYTE)x[2]]^
             ftable1[(BYTE)(x[3]>>8)]^
             ftable2[(BYTE)(x[0]>>16)]^
             ftable3[x[1]>>24];
        y[3]=a->fkey[k+3]^ftable[(BYTE)x[3]]^
             ftable1[(BYTE)(x[0]>>8)]^
             ftable2[(BYTE)(x[1]>>16)]^
             ftable3[x[2]>>24];
#else
        y[0]=a->fkey[k]^ftable[(BYTE)x[0]]^
             ROTL8(ftable[(BYTE)(x[1]>>8)])^
             ROTL16(ftable[(BYTE)(x[2]>>16)])^
             ROTL24(ftable[x[3]>>24]);
        y[1]=a->fkey[k+1]^ftable[(BYTE)x[1]]^
             ROTL8(ftable[(BYTE)(x[2]>>8)])^
             ROTL16(ftable[(BYTE)(x[3]>>16)])^
             ROTL24(ftable[x[0]>>24]);
        y[2]=a->fkey[k+2]^ftable[(BYTE)x[2]]^
             ROTL8(ftable[(BYTE)(x[3]>>8)])^
             ROTL16(ftable[(BYTE)(x[0]>>16)])^
             ROTL24(ftable[x[1]>>24]);
        y[3]=a->fkey[k+3]^ftable[(BYTE)x[3]]^
             ROTL8(ftable[(BYTE)(x[0]>>8)])^
             ROTL16(ftable[(BYTE)(x[1]>>16)])^
             ROTL24(ftable[x[2]>>24]);
#endif
        k+=4;
        t=x; x=y; y=t;      /* swap pointers */
    }

/* Last Round */ 
    
    y[0]=a->fkey[k]^(WORD)fbsub[(BYTE)x[0]]^
         ROTL8((WORD)fbsub[(BYTE)(x[1]>>8)])^
         ROTL16((WORD)fbsub[(BYTE)(x[2]>>16)])^
         ROTL24((WORD)fbsub[x[3]>>24]);
    y[1]=a->fkey[k+1]^(WORD)fbsub[(BYTE)x[1]]^
         ROTL8((WORD)fbsub[(BYTE)(x[2]>>8)])^
         ROTL16((WORD)fbsub[(BYTE)(x[3]>>16)])^
         ROTL24((WORD)fbsub[x[0]>>24]);
    y[2]=a->fkey[k+2]^(WORD)fbsub[(BYTE)x[2]]^
         ROTL8((WORD)fbsub[(BYTE)(x[3]>>8)])^
         ROTL16((WORD)fbsub[(BYTE)(x[0]>>16)])^
         ROTL24((WORD)fbsub[x[1]>>24]);
    y[3]=a->fkey[k+3]^(WORD)fbsub[(BYTE)x[3]]^
         ROTL8((WORD)fbsub[(BYTE)(x[0]>>8)])^
         ROTL16((WORD)fbsub[(BYTE)(x[1]>>16)])^
         ROTL24((WORD)fbsub[x[2]>>24]);

    for (i=j=0;i<NB;i++,j+=4)
    {
        unpack(y[i],&buff[j]);
        if (a->CBC) a->chain[i]=y[i];
        x[i]=y[i]=0;   /* clean up stack */
    }
    return;
}

void aes_decrypt(aes *a,char *buff)
{
    int i,j,k;
    WORD p[4],q[4],st[4],*x,*y,*t;


    for (i=j=0;i<NB;i++,j+=4)
    {
        p[i]=pack(&buff[j]);
        if (a->CBC) 
        {
            st[i]=a->chain[i];
            a->chain[i]=p[i];
        }
        p[i]^=a->rkey[i];
    }
    k=NB;
    x=p; y=q;

/* State alternates between p and q */
    for (i=1;i<a->Nr;i++)
    { /* Nr is number of rounds. May be odd. */
#ifdef LOTS_OF_MEMORY
        y[0]=a->rkey[k]^rtable[(BYTE)x[0]]^
             rtable1[(BYTE)(x[3]>>8)]^
             rtable2[(BYTE)(x[2]>>16)]^
             rtable3[x[1]>>24];
        y[1]=a->rkey[k+1]^rtable[(BYTE)x[1]]^
             rtable1[(BYTE)(x[0]>>8)]^
             rtable2[(BYTE)(x[3]>>16)]^
             rtable3[x[2]>>24];
        y[2]=a->rkey[k+2]^rtable[(BYTE)x[2]]^
             rtable1[(BYTE)(x[1]>>8)]^
             rtable2[(BYTE)(x[0]>>16)]^
             rtable3[x[3]>>24];
        y[3]=a->rkey[k+3]^rtable[(BYTE)x[3]]^
             rtable1[(BYTE)(x[2]>>8)]^
             rtable2[(BYTE)(x[1]>>16)]^
             rtable3[x[0]>>24];
#else
        y[0]=a->rkey[k]^rtable[(BYTE)x[0]]^
             ROTL8(rtable[(BYTE)(x[3]>>8)])^
             ROTL16(rtable[(BYTE)(x[2]>>16)])^
             ROTL24(rtable[x[1]>>24]);
        y[1]=a->rkey[k+1]^rtable[(BYTE)x[1]]^
             ROTL8(rtable[(BYTE)(x[0]>>8)])^
             ROTL16(rtable[(BYTE)(x[3]>>16)])^
             ROTL24(rtable[x[2]>>24]);
        y[2]=a->rkey[k+2]^rtable[(BYTE)x[2]]^
             ROTL8(rtable[(BYTE)(x[1]>>8)])^
             ROTL16(rtable[(BYTE)(x[0]>>16)])^
             ROTL24(rtable[x[3]>>24]);
        y[3]=a->rkey[k+3]^rtable[(BYTE)x[3]]^
             ROTL8(rtable[(BYTE)(x[2]>>8)])^
             ROTL16(rtable[(BYTE)(x[1]>>16)])^
             ROTL24(rtable[x[0]>>24]);
#endif
        k+=4;
        t=x; x=y; y=t;      /* swap pointers */
    }

/* Last Round */ 
    y[0]=a->rkey[k]^(WORD)rbsub[(BYTE)x[0]]^
         ROTL8((WORD)rbsub[(BYTE)(x[3]>>8)])^
         ROTL16((WORD)rbsub[(BYTE)(x[2]>>16)])^
         ROTL24((WORD)rbsub[x[1]>>24]);
    y[1]=a->rkey[k+1]^(WORD)rbsub[(BYTE)x[1]]^
         ROTL8((WORD)rbsub[(BYTE)(x[0]>>8)])^
         ROTL16((WORD)rbsub[(BYTE)(x[3]>>16)])^
         ROTL24((WORD)rbsub[x[2]>>24]);
    y[2]=a->rkey[k+2]^(WORD)rbsub[(BYTE)x[2]]^
         ROTL8((WORD)rbsub[(BYTE)(x[1]>>8)])^
         ROTL16((WORD)rbsub[(BYTE)(x[0]>>16)])^
         ROTL24((WORD)rbsub[x[3]>>24]);
    y[3]=a->rkey[k+3]^(WORD)rbsub[(BYTE)x[3]]^
         ROTL8((WORD)rbsub[(BYTE)(x[2]>>8)])^
         ROTL16((WORD)rbsub[(BYTE)(x[1]>>16)])^
         ROTL24((WORD)rbsub[x[0]>>24]);

    for (i=j=0;i<NB;i++,j+=4)
    {
        if (a->CBC)
        {
            y[i]^=st[i];
            st[i]=0;
        }
        unpack(y[i],&buff[j]);
        x[i]=y[i]=0;   /* clean up stack */
    }
    return;
}

void aes_end(aes *a)
{ /* clean up */
    int i;
    for (i=0;i<NB*(a->Nr+1);i++)
        a->fkey[i]=a->rkey[i]=0;
    for (i=0;i<NB;i++) a->chain[i]=0;
    mr_free(a->rkey);
    mr_free(a->fkey);
}

/*
int main()
{ 
    int i,j,nk;
    aes a;
    BYTE y,x,m;
    char key[32];
    char block[16];
    char iv[16];
    for (i=0;i<32;i++) key[i]=0;
    key[0]=1;
    for (i=0;i<16;i++) iv[i]=i;
    for (i=0;i<16;i++) block[i]=i;

    for (nk=16;nk<=32;nk+=8)
    {  
        printf("\nKey Size= %d bits\n",nk*8);
        if (!aes_init(&a,nk,key,iv))
        {
            printf("Failed to Initialize\n");
            return 0;
        }

        printf("Plain=   ");
        for (i=0;i<4*NB;i++) printf("%2x",block[i]);
        printf("\n");
        aes_encrypt(&a,block);
        printf("Encrypt= ");
        for (i=0;i<4*NB;i++) printf("%2x",(unsigned char)block[i]);
        printf("\n");
        aes_reset(&a,iv);
        aes_decrypt(&a,block);
        printf("Decrypt= ");
        for (i=0;i<4*NB;i++) printf("%2x",block[i]);
        printf("\n");

        aes_end(&a);
    }
    return 0;
}
*/

