/*
 *   MIRACL elliptic curve routines
 *   mrcurve.c
 *
 *   Assumes Weierstrass equation y^2 = x^3 + Ax + B
 *   See IEEE P1363 Draft Standard 
 *
 *   Uses Montgomery's representation internally
 *
 *   Works particularly well with fixed length Comba multiplier
 *   e.g. #define MR_COMBA 5 for 5x32 = 160 bit modulus
 *        on 32-bit computer
 *
 *   Copyright (c) 1997-1999 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

#define mr_abs(x)  ((x)<0? (-(x)) : (x))

#define MR_EPOINT_GENERAL    0
#define MR_EPOINT_NORMALIZED 1
#define MR_EPOINT_INFINITY   2

extern miracl *mr_mip;

void ecurve_init(big a,big b,big p,int type)
{ /* Initialize the active ecurve    *
   * Asize indicate size of A        *
   * Bsize indicate size of B        */
    int as;
    if (mr_mip->ERNUM) return;

    MR_IN(93)

    prepare_monty(p);

    mr_mip->Asize=size(a);
    if (mr_abs(mr_mip->Asize)==mr_mip->TOOBIG)
    {
        if (mr_mip->Asize>=0)
        { /* big positive number - check it isn't minus something small */
           copy(a,mr_mip->w1);
           divide(mr_mip->w1,p,p);
           subtract(p,mr_mip->w1,mr_mip->w1);
           as=size(mr_mip->w1);
           if (as<mr_mip->TOOBIG) mr_mip->Asize=-as;
           else
           {
               if (mr_mip->A==NULL) mr_mip->A=mirvar(0);
               nres(a,mr_mip->A);
           }
        }
        else
        {
               if (mr_mip->A==NULL) mr_mip->A=mirvar(0);
               nres(a,mr_mip->A);
        }
    }
    mr_mip->Bsize=size(b);
    if (mr_abs(mr_mip->Bsize)==mr_mip->TOOBIG)
    {
        if (mr_mip->B==NULL) mr_mip->B=mirvar(0);
        nres(b,mr_mip->B);
    }
    mr_mip->coord=type;
    MR_OUT
    return;
}

epoint* epoint_init(void)
{ /* initialise epoint to point at infinity. */
    epoint *p;
    if (mr_mip->ERNUM) return NULL;

    MR_IN(96)

    p=mr_alloc(1,sizeof(epoint));
    p->X=mirvar(0);
    p->Y=mirvar(0);
    p->Z=NULL;            /* only initialise when written to */
    p->marker=MR_EPOINT_INFINITY;

    MR_OUT

    return p;
}

void epoint_free(epoint *p)
{ /* clean up point */
    mirkill(p->X);
    mirkill(p->Y);
    if (p->Z!=NULL) mirkill(p->Z);
    mr_free(p);
}        

BOOL epoint_set(big x,big y,int cb,epoint *p)
{ /* initialise a point on active ecurve            *
   * if x or y == NULL, set to point at infinity    *
   * if x==y, a y co-ordinate is calculated - if    *
   * possible - and cb suggests LSB 0/1  of y       *
   * (which "decompresses" x). Otherwise, check     *
   * validity of given (x,y) point, ignoring cb.    *
   * Returns TRUE for valid point, otherwise FALSE. */
  
    BOOL valid;

    if (mr_mip->ERNUM) return FALSE;

    MR_IN(97)

    if (x==NULL || y==NULL)
    {
        convert(1,mr_mip->w1);
        nres(mr_mip->w1,p->X);
        nres(mr_mip->w1,p->Y);
        p->marker=MR_EPOINT_INFINITY;
        MR_OUT
        return TRUE;
    }

/* find x^3+Ax+B */
    nres(x,p->X);
    nres_modmult(p->X,p->X,mr_mip->w3);
    nres_modmult(mr_mip->w3,p->X,mr_mip->w3);
    if (mr_abs(mr_mip->Asize)==mr_mip->TOOBIG)
        nres_modmult(p->X,mr_mip->A,mr_mip->w1);
    else
        nres_premult(p->X,mr_mip->Asize,mr_mip->w1);
    nres_modadd(mr_mip->w3,mr_mip->w1,mr_mip->w3);
    if (mr_abs(mr_mip->Bsize)==mr_mip->TOOBIG)
        nres_modadd(mr_mip->w3,mr_mip->B,mr_mip->w3);
    else
    {
        convert(mr_mip->Bsize,mr_mip->w1);
        nres(mr_mip->w1,mr_mip->w1);
        nres_modadd(mr_mip->w3,mr_mip->w1,mr_mip->w3);
    }
    valid=FALSE;
    if (x!=y)
    { /* compare with y^2 */
        nres(y,p->Y);
        nres_modmult(p->Y,p->Y,mr_mip->w1);
        if (compare(mr_mip->w1,mr_mip->w3)==0) valid=TRUE;
    }
    else
    { /* no y supplied - calculate one. Find square root */

        valid=nres_sqroot(mr_mip->w3,p->Y);

    /* check LSB - have we got the right root? */
        redc(p->Y,mr_mip->w1);
        if (remain(mr_mip->w1,2)!=cb) 
            mr_psub(mr_mip->modulus,p->Y,p->Y);

    } 
    if (valid)
    {
        if (p->Z!=NULL)
        {
            convert(1,mr_mip->w1);
            nres(mr_mip->w1,p->Z);
        }
        p->marker=MR_EPOINT_NORMALIZED;
        MR_OUT
        return TRUE;
    }
    MR_OUT
    return FALSE;
}

int epoint_get(epoint* p,big x,big y)
{ /* Get point co-ordinates in affine, normal form       *
   * (converted from projective, Montgomery form)        *
   * if x==y, supplies x only. Return value is Least     *
   * Significant Bit of y (useful for point compression) */

    int lsb;

    if (p->marker==MR_EPOINT_INFINITY)
    {
        zero(x);
        zero(y);
        return 0;
    }
    if (mr_mip->ERNUM) return 0;

    MR_IN(98)

    if (epoint_norm(p)<0) 
    { /* not possible ! */
        MR_OUT
        return (-1);
    }

    redc(p->X,x);
    redc(p->Y,mr_mip->w1);

    if (x!=y) copy(mr_mip->w1,y);
    lsb=remain(mr_mip->w1,2); 
    MR_OUT
    return lsb;
}

BOOL epoint_norm(epoint *p)
{ /* normalise a point */
    
    if (mr_mip->coord==MR_AFFINE) return TRUE;
    if (p->marker!=MR_EPOINT_GENERAL) return TRUE;

    if (mr_mip->ERNUM) return FALSE;

    MR_IN(117)

    convert(1,mr_mip->w8);
    nres(mr_mip->w8,mr_mip->w8);  

    if (nres_moddiv(mr_mip->w8,p->Z,mr_mip->w8)>1) /* 1/Z  */
    {
        MR_OUT
        return FALSE;
    }
    
    nres_modmult(mr_mip->w8,mr_mip->w8,mr_mip->w1);/* 1/ZZ */
    nres_modmult(p->X,mr_mip->w1,p->X);            /* X/ZZ */
    nres_modmult(mr_mip->w1,mr_mip->w8,mr_mip->w1);/* 1/ZZZ */
    nres_modmult(p->Y,mr_mip->w1,p->Y);            /* Y/ZZZ */

    convert(1,mr_mip->w8);
    nres(mr_mip->w8,p->Z);
   
    p->marker=MR_EPOINT_NORMALIZED;
    MR_OUT
    return TRUE;
}

/*

static BOOL epoint_multi_norm(int m,epoint **p)
{ 
    int i;
    big x[11],w[11];
    for (i=0;i<m;i++) 
    {  
        w[i]=mirvar(0);
        x[i]=p[i]->Z;
    }
    if (!nres_multi_inverse(m,x,w)) 
    {
       for (i=0;i<m;i++) mirkill(w[i]); 
       return FALSE;
    }

    convert(1,mr_mip->w1);
    nres(mr_mip->w1,mr_mip->w1);

    for (i=0;i<m;i++)
    {
        nres_modmult(w[i],w[i],mr_mip->w2);      
        nres_modmult(w[i],mr_mip->w2,mr_mip->w3);
        nres_modmult(p[i]->X,mr_mip->w2,p[i]->X);
        nres_modmult(p[i]->Y,mr_mip->w3,p[i]->Y);
        copy(mr_mip->w1,p[i]->Z);
        p[i]->marker=MR_EPOINT_NORMALIZED;
        mirkill(w[i]);
    }
    return TRUE;
}

*/

static void ecurve_double(epoint *p)
{ /* double epoint on active ecurve */

    if (p->marker==MR_EPOINT_INFINITY) 
    { /* 2 times infinity == infinity ! */
        return;
    }

    if (mr_mip->coord==MR_AFFINE)
    {
        if (size(p->Y)==0) 
        { /* set to point at infinity */
            epoint_set(NULL,NULL,0,p);
            return;
        }
        nres_modmult(p->X,p->X,mr_mip->w8);    /* w8=x^2   */
        nres_premult(mr_mip->w8,3,mr_mip->w8); /* w8=3*x^2 */
        if (mr_abs(mr_mip->Asize) == mr_mip->TOOBIG)
            nres_modadd(mr_mip->w8,mr_mip->A,mr_mip->w8);
        else
        {
            convert(mr_mip->Asize,mr_mip->w2);
            nres(mr_mip->w2,mr_mip->w2);
            nres_modadd(mr_mip->w8,mr_mip->w2,mr_mip->w8);
        }                                     /* w8=3*x^2+A */
        nres_premult(p->Y,2,mr_mip->w6);      /* w6=2y */
        nres_moddiv(mr_mip->w8,mr_mip->w6,mr_mip->w8);  /* w8 is slope m */

        nres_modmult(mr_mip->w8,mr_mip->w8,mr_mip->w7); /* w7=m^2 */
        nres_premult(p->X,2,mr_mip->w1);
        nres_modsub(mr_mip->w7,mr_mip->w1,mr_mip->w1); /* w1=m^2-2x */
        
        nres_modsub(p->X,mr_mip->w1,mr_mip->w2);
        nres_modmult(mr_mip->w2,mr_mip->w8,mr_mip->w2);
        nres_modsub(mr_mip->w2,p->Y,p->Y);
        copy(mr_mip->w1,p->X);
        
        return;
    }

    if (p->Z==NULL)
    {
        p->Z=mirvar(1);
        nres(p->Z,p->Z);
    }

    if (size(p->Y)==0)
    { /* set to point at infinity */
        epoint_set(NULL,NULL,0,p);
        return;
    }
    convert(1,mr_mip->w1);
    if (p->marker==MR_EPOINT_NORMALIZED) nres(mr_mip->w1,mr_mip->w2);
    else nres_modmult(p->Z,p->Z,mr_mip->w2);
    if (mr_abs(mr_mip->Asize) < mr_mip->TOOBIG)
    {
        if (mr_mip->Asize==(-3))
        { /* a is -3. Goody. */
            nres_modsub(p->X,mr_mip->w2,mr_mip->w3);
            nres_modadd(p->X,mr_mip->w2,mr_mip->w4);
            nres_modmult(mr_mip->w3,mr_mip->w4,mr_mip->w3);
            nres_modadd(mr_mip->w3,mr_mip->w3,mr_mip->w4);
            nres_modadd(mr_mip->w4,mr_mip->w3,mr_mip->w4);
        }
        else
        { /* a is small */
            if (mr_mip->Asize!=0)
            { /* a is non zero! */
                nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w3);
                nres_premult(mr_mip->w3,mr_mip->Asize,mr_mip->w3);
            }
            nres_modmult(p->X,p->X,mr_mip->w1);
            nres_modadd(mr_mip->w1,mr_mip->w1,mr_mip->w4);
            nres_modadd(mr_mip->w4,mr_mip->w1,mr_mip->w4);
            if (mr_mip->Asize!=0) nres_modadd(mr_mip->w4,mr_mip->w3,mr_mip->w4);
        }
    }
    else
    { /* a is not special */
        nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w3);
        nres_modmult(mr_mip->w3,mr_mip->A,mr_mip->w3);
        nres_modmult(p->X,p->X,mr_mip->w1);
        nres_modadd(mr_mip->w1,mr_mip->w1,mr_mip->w4);
        nres_modadd(mr_mip->w4,mr_mip->w1,mr_mip->w4);
        nres_modadd(mr_mip->w4,mr_mip->w3,mr_mip->w4);        
    }
    nres_modmult(p->Y,p->Z,p->Z);
    nres_modadd(p->Z,p->Z,p->Z);
    nres_modmult(p->Y,p->Y,mr_mip->w2);
    nres_modmult(p->X,mr_mip->w2,mr_mip->w3);
    nres_modadd(mr_mip->w3,mr_mip->w3,mr_mip->w3);
    nres_modadd(mr_mip->w3,mr_mip->w3,mr_mip->w3);
    nres_modmult(mr_mip->w4,mr_mip->w4,mr_mip->w1);
    nres_modadd(mr_mip->w3,mr_mip->w3,p->X);
    nres_modsub(mr_mip->w1,p->X,p->X);
    nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w2);
    nres_modadd(mr_mip->w2,mr_mip->w2,mr_mip->w2);
    nres_modadd(mr_mip->w2,mr_mip->w2,mr_mip->w2);
    nres_modadd(mr_mip->w2,mr_mip->w2,mr_mip->w2);
    nres_modsub(mr_mip->w3,p->X,mr_mip->w3);
    nres_modmult(mr_mip->w3,mr_mip->w4,mr_mip->w3);
    nres_modsub(mr_mip->w3,mr_mip->w2,p->Y);
  
    p->marker=MR_EPOINT_GENERAL;
}
   
static BOOL ecurve_padd(epoint *p,epoint *pa)
{ /* primitive add two epoints on the active ecurve - pa+=p;   *
   * note that if p is normalized, its Z coordinate isn't used */
 
    if (mr_mip->coord==MR_AFFINE)
    {
        nres_modsub(p->Y,pa->Y,mr_mip->w8);
        nres_modsub(p->X,pa->X,mr_mip->w6);
        if (size(mr_mip->w6)==0) 
        { /* divide by 0 */
            if (size(mr_mip->w8)==0) 
            { /* should have doubled ! */
                return FALSE; 
            }
            else
            { /* point at infinity */
                epoint_set(NULL,NULL,0,pa);
                return TRUE;
            }
        }
        nres_moddiv(mr_mip->w8,mr_mip->w6,mr_mip->w8);

        nres_modmult(mr_mip->w8,mr_mip->w8,mr_mip->w7); /* w7=m^2 */
        nres_modsub(mr_mip->w7,p->X,mr_mip->w1); /* w1=m^2-x1-x2 */
        nres_modsub(mr_mip->w1,pa->X,mr_mip->w1);
        

        nres_modsub(pa->X,mr_mip->w1,mr_mip->w2);
        nres_modmult(mr_mip->w2,mr_mip->w8,mr_mip->w2);
        nres_modsub(mr_mip->w2,pa->Y,pa->Y);
        copy(mr_mip->w1,pa->X);

        return TRUE;
    }

    if (pa->Z==NULL) 
    {
        pa->Z=mirvar(1);
        nres(pa->Z,pa->Z);
    }
    if (p->marker!=MR_EPOINT_NORMALIZED)    
    {
        nres_modmult(p->Z,p->Z,mr_mip->w6);
        nres_modmult(pa->X,mr_mip->w6,mr_mip->w1);
        nres_modmult(mr_mip->w6,p->Z,mr_mip->w6);
        nres_modmult(pa->Y,mr_mip->w6,mr_mip->w2);
    }
    else
    {
        copy(pa->X,mr_mip->w1);
        copy(pa->Y,mr_mip->w2);
    }
    nres_modmult(pa->Z,pa->Z,mr_mip->w6);
    nres_modmult(p->X,mr_mip->w6,mr_mip->w4);
    nres_modmult(mr_mip->w6,pa->Z,mr_mip->w6);
    nres_modmult(p->Y,mr_mip->w6,mr_mip->w5);
    nres_modsub(mr_mip->w1,mr_mip->w4,mr_mip->w1);
    nres_modsub(mr_mip->w2,mr_mip->w5,mr_mip->w2);

    if (size(mr_mip->w1)==0)
    {
        if (size(mr_mip->w2)==0)
        { /* should have doubled ! */
           return FALSE; 
        }
        else
        { /* point at infinity */
            epoint_set(NULL,NULL,0,pa);
            return TRUE;
        }
    }
    nres_modadd(mr_mip->w4,mr_mip->w4,mr_mip->w6);
    nres_modadd(mr_mip->w1,mr_mip->w6,mr_mip->w4);
    nres_modadd(mr_mip->w5,mr_mip->w5,mr_mip->w6);
    nres_modadd(mr_mip->w2,mr_mip->w6,mr_mip->w5);
    
    if (p->marker!=MR_EPOINT_NORMALIZED)
    { 
        nres_modmult(pa->Z,p->Z,mr_mip->w3);
        nres_modmult(mr_mip->w3,mr_mip->w1,pa->Z);
    }
    else nres_modmult(pa->Z,mr_mip->w1,pa->Z);
    nres_modmult(mr_mip->w1,mr_mip->w1,mr_mip->w6);
    nres_modmult(mr_mip->w1,mr_mip->w6,mr_mip->w1);
    nres_modmult(mr_mip->w6,mr_mip->w4,mr_mip->w6);
    nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w4);

    nres_modsub(mr_mip->w4,mr_mip->w6,pa->X);

    nres_modsub(mr_mip->w6,pa->X,mr_mip->w6);
    nres_modsub(mr_mip->w6,pa->X,mr_mip->w6);
    nres_modmult(mr_mip->w2,mr_mip->w6,mr_mip->w2);
    nres_modmult(mr_mip->w1,mr_mip->w5,mr_mip->w1);
    nres_modsub(mr_mip->w2,mr_mip->w1,mr_mip->w5);

/* divide by 2 */
    if (subdiv(mr_mip->w5,2,pa->Y)!=0)
    {
        mr_padd(mr_mip->modulus,mr_mip->w5,pa->Y);
        subdiv(pa->Y,2,pa->Y);
    }
    pa->marker=MR_EPOINT_GENERAL;
    return TRUE;                                      
}

void epoint_copy(epoint *a,epoint *b)
{   
    if (a==b) return;
    copy(a->X,b->X);
    copy(a->Y,b->Y);
    if (a->Z==NULL)
    {
         if (b->Z!=NULL)
         {
            convert(1,mr_mip->w2);
            nres(mr_mip->w2,b->Z);
         }
    }
    else 
    {
        if (b->Z==NULL)
        {
           if (a->marker==MR_EPOINT_GENERAL) 
           {
               b->Z=mirvar(0);
               copy(a->Z,b->Z);
           } 
        }
        else copy(a->Z,b->Z);
    }
    b->marker=a->marker;
    return;
}

BOOL epoint_comp(epoint *a,epoint *b)
{
    int ia,ib;
    if (mr_mip->ERNUM) return FALSE;
    if (a==b) return TRUE;

    MR_IN(105)

    ia=epoint_get(a,mr_mip->w9,mr_mip->w9);
    ib=epoint_get(b,mr_mip->w10,mr_mip->w10);

    MR_OUT
    if (ia==ib && compare(mr_mip->w9,mr_mip->w10)==0) return TRUE;
    return FALSE;
}

void ecurve_add(epoint *p,epoint *pa)
{  /* pa=pa+p; */
    if (mr_mip->ERNUM) return;

    MR_IN(94)

    if (p==pa) 
    {
        ecurve_double(pa);
        MR_OUT
        return;
    }
    if (pa->marker==MR_EPOINT_INFINITY)
    {
        epoint_copy(p,pa);
        MR_OUT 
        return;
    }
    if (p->marker==MR_EPOINT_INFINITY) 
    {
        MR_OUT
        return;
    }
    if (!ecurve_padd(p,pa)) ecurve_double(pa);
    MR_OUT
}

void ecurve_sub(epoint *p,epoint *pa)
{
    if (mr_mip->ERNUM) return;

    MR_IN(104)

    if (p==pa)
    {
        epoint_set(NULL,NULL,0,pa);
        MR_OUT
        return;
    } 
    if (p->marker==MR_EPOINT_INFINITY) 
    {
        MR_OUT
        return;
    }

    mr_psub(mr_mip->modulus,p->Y,p->Y);
    ecurve_add(p,pa);
    mr_psub(mr_mip->modulus,p->Y,p->Y);

    MR_OUT
}


static int mr_naf_window(big x,big x3,int i,int *nbs,int *nzs)
{ /* returns sliding window value, max of 5 bits           *
   * starting at i-th bit of x. nbs is number of bits      *
   * processed. nzs is number of additional trailing       *    
   * zeros detected. x and x3 (which is 3*x) are           *
   * combined to produce the NAF (non-adjacent form)       *
   * So if x=11011(27) and x3 is 1010001, the LSB is       *
   * ignored and the value 100T0T (32-4-1=27) processed,   *
   * where T is -1. Note x.P = (3x-x)/2.P. This value will *
   * return +7, with nbs=4 and nzs=1, having stopped after *
   * the first 4 bits. Note in an NAF non-zero elements    *
   * are never side by side, so 10T10T won't happen        */

    int nb,j,r,w;
    BOOL last;
    w=5;

 /* get first bit */
    nb=mr_testbit(x3,i)-mr_testbit(x,i);

    *nbs=1;
    *nzs=0;
    if (nb==0) return 0;
    last=FALSE;
    if (i<=w) 
    {
        w=i;
        last=TRUE;
    }

    if (nb>0) r=1;
    else      r=(-1);

    for (j=i-1;j>i-w;j--)
    { /* scan the bits */
        (*nbs)++;
        r*=2;
        nb=mr_testbit(x3,j)-mr_testbit(x,j);
        if (nb==0) continue;
        if (nb>0) r+=1;
        if (nb<0) r-=1;
           
    } 
    if (!last && r%2!=0) (*nzs)++;
    while (r%2==0)
    { /* remove trailing zeros */
        r/=2;
        (*nzs)++;
        (*nbs)--;
    }     
    return r;
}

void ecurve_mult(big e,epoint *pa,epoint *pt)
{ /* pt=e*pa; */
    int i,j,n,ch,ce,nb,nbs,nzs;
    epoint *p;
    epoint *table[11];
    if (mr_mip->ERNUM) return;

    MR_IN(95)

    if (size(e)==0) 
    { /* multiplied by 0 */
        epoint_set(NULL,NULL,0,pt);
        MR_OUT
        return;
    }
    copy(e,mr_mip->w9);
    epoint_norm(pa);
    epoint_copy(pa,pt);

    if (size(mr_mip->w9)<0)
    { /* pt = -pt */
        negate(mr_mip->w9,mr_mip->w9);
        mr_psub(mr_mip->modulus,pt->Y,pt->Y);
    }

    if (size(mr_mip->w9)==1)
    { 
        MR_OUT
        return;
    }

    premult(mr_mip->w9,3,mr_mip->w10);      /* h=3*e */
    p=epoint_init();
    epoint_copy(pt,p);

    if (mr_mip->base==mr_mip->base2)
    {
        table[0]=epoint_init();
        epoint_copy(p,table[0]);
        ecurve_double(p);

        for (i=1;i<=10;i++)
        { /* precomputation */
            table[i]=epoint_init();
            epoint_copy(table[i-1],table[i]);
            ecurve_add(p,table[i]);
        }

  /* note that normalising this table doesn't really help */
        nb=logb2(mr_mip->w10);

        for (i=nb-2;i>=1;)
        { /* add/subtract */
            if (mr_mip->user!=NULL) (*mr_mip->user)();
            n=mr_naf_window(mr_mip->w9,mr_mip->w10,i,&nbs,&nzs);
            for (j=0;j<nbs;j++)
                ecurve_double(pt);
            if (n>0) ecurve_add(table[n/2],pt);
            if (n<0) ecurve_sub(table[(-n)/2],pt);
            i-=nbs;
            if (nzs)
            {
                for (j=0;j<nzs;j++) ecurve_double(pt);
                i-=nzs;
            }
        }
        for (i=10;i>=0;i--) epoint_free(table[i]);
    }
    else
    { 
        expint(2,logb2(mr_mip->w10)-1,mr_mip->w11);
        mr_psub(mr_mip->w10,mr_mip->w11,mr_mip->w10);
        subdiv(mr_mip->w11,2,mr_mip->w11);
        while (size(mr_mip->w11) > 1)
        { /* add/subtract method */
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            ecurve_double(pt);
            ce=compare(mr_mip->w9,mr_mip->w11); /* e(i)=1? */
            ch=compare(mr_mip->w10,mr_mip->w11); /* h(i)=1? */
            if (ch>=0) 
            {  /* h(i)=1 */
                if (ce<0) ecurve_add(p,pt);
                mr_psub(mr_mip->w10,mr_mip->w11,mr_mip->w10);
            }
            if (ce>=0) 
            {  /* e(i)=1 */
                if (ch<0) ecurve_sub(p,pt);
                mr_psub(mr_mip->w9,mr_mip->w11,mr_mip->w9);  
            }
            subdiv(mr_mip->w11,2,mr_mip->w11);
        }
    }
    epoint_free(p);
    MR_OUT
}

void ecurve_multn(int n,big *y,epoint **x,epoint *w)
{ /* pt=e[o]*p[0]+e[1]*p[1]+ .... e[n-1]*p[n-1]   */
    int i,j,k,m,nb,ea;
    epoint **G;
    if (mr_mip->ERNUM) return;

    MR_IN(113)

    m=1<<n;
    G=(epoint **)mr_alloc(m,sizeof(epoint*));

    for (i=0,k=1;i<n;i++)
    {
        for (j=0; j < (1<<i) ;j++)
        {
            G[k]=epoint_init();
            epoint_copy(x[i],G[k]);
            if (j!=0) ecurve_add(G[j],G[k]);
            k++;
        }
    }

    nb=0;
    for (j=0;j<n;j++) if ((k=logb2(y[j])) > nb) nb=k;

    epoint_set(NULL,NULL,0,w);            /* w=0 */
    
    if (mr_mip->base==mr_mip->base2)
    {
        for (i=nb-1;i>=0;i--)
        {
            if (mr_mip->user!=NULL) (*mr_mip->user)();
            ea=0;
            k=1;
            for (j=0;j<n;j++)
            {
                if (mr_testbit(y[j],i)) ea+=k;
                k<<=1;
            }
            ecurve_double(w);
            if (ea!=0) ecurve_add(G[ea],w);
        }    
    }
    else mr_berror(MR_ERR_NOT_SUPPORTED);

    for (i=1;i<m;i++) epoint_free(G[i]);
    mr_free(G);
    MR_OUT
}

void ecurve_mult2(big e,epoint *p,big ea,epoint *pa,epoint *pt)
{ /* pt=e*p+ea*pa; */
    int e1,h1,e2,h2;
    epoint *p1,*p2,*ps,*pd;
    if (mr_mip->ERNUM) return;

    MR_IN(103)

    if (size(e)==0) 
    {
        ecurve_mult(ea,pa,pt);
        MR_OUT
        return;
    }

    p2=epoint_init();
    epoint_norm(pa);
    epoint_copy(pa,p2);
    copy(ea,mr_mip->w9);
    if (size(mr_mip->w9)<0)
    { /* p2 = -p2 */
        negate(mr_mip->w9,mr_mip->w9);
        mr_psub(mr_mip->modulus,p2->Y,p2->Y);
    }
    premult(mr_mip->w9,3,mr_mip->w10);      /* 3*ea */

    p1=epoint_init();
    epoint_norm(p);
    epoint_copy(p,p1);
    copy(e,mr_mip->w12);
    if (size(mr_mip->w12)<0)
    { /* p1= -p1 */
        negate(mr_mip->w12,mr_mip->w12);
        mr_psub(mr_mip->modulus,p1->Y,p1->Y);
    }
    premult(mr_mip->w12,3,mr_mip->w13);    /* 3*e */

    epoint_set(NULL,NULL,0,pt);            /* pt=0 */

    if (compare(mr_mip->w10,mr_mip->w13)>=0)
         expint(2,logb2(mr_mip->w10)-1,mr_mip->w11);
    else expint(2,logb2(mr_mip->w13)-1,mr_mip->w11);

    ps=epoint_init();
    pd=epoint_init();

    epoint_copy(p1,ps);
    ecurve_add(p2,ps);                    /* ps=p1+p2 */
    epoint_copy(p1,pd);
    ecurve_sub(p2,pd);                    /* pd=p1-p2 */
    epoint_norm(ps);
    epoint_norm(pd);
    while (size(mr_mip->w11) > 1)
    { /* add/subtract method */
        if (mr_mip->user!=NULL) (*mr_mip->user)();

        ecurve_double(pt);

        e1=h1=e2=h2=0;
        if (compare(mr_mip->w9,mr_mip->w11)>=0)
        { /* e1(i)=1? */
            e2=1;  
            mr_psub(mr_mip->w9,mr_mip->w11,mr_mip->w9);
        }
        if (compare(mr_mip->w10,mr_mip->w11)>=0)
        { /* h1(i)=1? */
            h2=1;  
            mr_psub(mr_mip->w10,mr_mip->w11,mr_mip->w10);
        } 
        if (compare(mr_mip->w12,mr_mip->w11)>=0)
        { /* e2(i)=1? */
            e1=1;   
            mr_psub(mr_mip->w12,mr_mip->w11,mr_mip->w12);
        }
        if (compare(mr_mip->w13,mr_mip->w11)>=0) 
        { /* h2(i)=1? */
            h1=1;  
            mr_psub(mr_mip->w13,mr_mip->w11,mr_mip->w13);
        }

        if (e1!=h1)
        {
            if (e2==h2)
            {
                if (h1==1) ecurve_add(p1,pt);
                else       ecurve_sub(p1,pt);
            }
            else
            {
                if (h1==1)
                {
                    if (h2==1) ecurve_add(ps,pt);
                    else       ecurve_add(pd,pt);
                }
                else
                {
                    if (h2==1) ecurve_sub(pd,pt);
                    else       ecurve_sub(ps,pt);
                }
            }
        }
        else if (e2!=h2)
        {
            if (h2==1) ecurve_add(p2,pt);
            else       ecurve_sub(p2,pt);
        }

        subdiv(mr_mip->w11,2,mr_mip->w11);
    }
    epoint_free(p1);
    epoint_free(p2);
    epoint_free(ps);
    epoint_free(pd);
    MR_OUT
}

