/*
 *   MIRACL fast fourier multiplication routine, using 3 prime method.
 *   mrfast.c  - only faster for very high precision multiplication
 *               of numbers > about 4096 bits  (see below)
 *   See "The Fast Fourier Transform in a Finite Field"  by J.M. Pollard,
 *   Mathematics of Computation, Vol. 25, No. 114, April 1971, pp365-374
 *   Also Knuth Vol. 2, Chapt. 4.3.3
 *
 *   Takes time preportional to 9+15N+9N.lg(N) to multiply two different 
 *   N digit numbers. This reduces to 6+18N+6N.lg(N) when squaring.
 *   The classic method takes N.N and N(N+1)/2 respectively
 *
 *   Fast Polynomial arithmetic 
 * 
 *   See "A New Polynomial Factorisation Algorithm and its Implementation"
 *   by Victor Shoup, Jl. Symbolic Computation 1996
 *   Uses FFT method for fast arithmetic of large degree polynomials
 *
 *   Copyright (c) 1988-1999 Shamus Software Ltd.
 */

#include <stdio.h>
#include <stdlib.h>
#include <miracl.h>

extern miracl *mr_mip;

int mr_fft_init(int logn,big m1,big m2,BOOL cr)
{ /* logn is number of bits, m1 and m2 are the largest elements 
     that will arise in each number to be multiplied */
    mr_small kk;
    int i,j,newn,N,pr;
    mr_utype p,proot;

    newn=(1<<logn);
    multiply(m1,m2,mr_mip->w1);
    premult(mr_mip->w1,2*newn+1,mr_mip->w1);
    kk=((mr_small)1<<(MIRACL-2-logn));
    if (mr_mip->base!=0) while (kk*newn>mr_mip->base) kk>>=1;
    pr=0;
    while (size(mr_mip->w1)>0)
    { /* find out how many primes will be needed */
        do
        { 
            kk--;
            p=kk*newn+1;
        } while(spmd((mr_small)2,p-1,p)!=1);
        mr_sdiv(mr_mip->w1,p,mr_mip->w1);
        pr++;
    }
/* if nothing has changed, don't recalculate */
    if (logn<=mr_mip->logN && pr==mr_mip->nprimes) return pr;
    fft_reset();


    mr_mip->prime=(mr_utype *)mr_alloc(pr,sizeof(mr_utype));
    mr_mip->inverse=(mr_utype *)mr_alloc(pr,sizeof(mr_utype));
    mr_mip->roots=(mr_utype**)mr_alloc(pr,sizeof(mr_utype *));
    mr_mip->t= (mr_utype **)mr_alloc(pr,sizeof(mr_utype *));
    mr_mip->cr=(mr_utype *)mr_alloc(pr,sizeof(mr_utype));
    mr_mip->w=(mr_utype *)mr_alloc(newn,sizeof(mr_utype));

    kk=((mr_small)1<<(MIRACL-2-logn));
    if (mr_mip->base!=0) while (kk*newn>mr_mip->base) kk>>=1;
    for (i=0;i<pr;i++)
    { /* find the primes again */
        mr_mip->roots[i]=(mr_utype *)mr_alloc(newn,sizeof(mr_utype));
        mr_mip->t[i]=(mr_utype *)mr_alloc(newn,sizeof(mr_utype));

        do
        { 
            kk--;
            p=kk*newn+1;
        } while(spmd((mr_small)2,p-1,p)!=1);

        proot=p-1;
        for (j=1;j<logn;j++) proot=sqrmp(proot,p); 
        mr_mip->roots[i][0]=proot;   /* build residue table */
        for (j=1;j<newn;j++) mr_mip->roots[i][j]=smul(mr_mip->roots[i][j-1],proot,p);
        mr_mip->inverse[i]=invers((mr_small)newn,p);
        mr_mip->prime[i]=p;
    }

    mr_mip->logN=logn;

    mr_mip->nprimes=pr;
/* set up chinese remainder structure */
    if (cr)
        if (!scrt_init(&mr_mip->chin,pr,mr_mip->prime)) fft_reset();
    return pr;
}

void fft_reset()
{ /* reclaim any space used by FFT */
    int i;

    if (mr_mip->degree!=0)
    { /* clear any precomputed tables */
        for (i=0;i<mr_mip->nprimes;i++)
        {
            mr_free(mr_mip->s1[i]);
            mr_free(mr_mip->s2[i]);
        }
        mr_free(mr_mip->s1);
        mr_free(mr_mip->s2);
        mr_mip->degree=0;
    }

    if (mr_mip->logN!=0)
    { /* clear away old stuff */
        for (i=0;i<mr_mip->nprimes;i++) 
        {
            mr_free(mr_mip->roots[i]);
            mr_free(mr_mip->t[i]);
        }
        mr_free(mr_mip->w);
        mr_free(mr_mip->cr);
        mr_free(mr_mip->t);
        mr_free(mr_mip->roots);
        mr_free(mr_mip->inverse);
        mr_free(mr_mip->prime);
        mr_mip->nprimes=0;
        mr_mip->logN=0;
    }
/* clear CRT structure */
    if (mr_mip->chin.NP!=0) scrt_end(&mr_mip->chin);
}

void mr_dif_fft(int logn,int pr,mr_utype *data)
{ /* decimate-in-frequency fourier transform */
    int mmax,m,j,k,istep,i,ii,jj,r,newn,offset;
    mr_utype w,temp,prime,*roots;
#ifdef MR_NOASM
    mr_large dble;
#endif
    prime=mr_mip->prime[pr];
    roots=mr_mip->roots[pr];

    newn=(1<<logn);
    offset=(mr_mip->logN-logn);
    mmax=newn;
    for (k=0;k<logn;k++) {
        istep=mmax;
        mmax>>=1; 
        ii=newn;
        jj=newn/istep;
        ii-=jj;
        for (i=0;i<newn;i+=istep)
        { /* special case root=1 */ 
            j=i+mmax;
            temp=data[i]-data[j];
            if (temp<0) temp+=prime;
            data[i]+=data[j];
            if (data[i]>prime) data[i]-=prime;
            data[j]=temp;
        }
        for (m=1;m<mmax;m++) {

            w=roots[(ii<<offset)-1];
            ii-=jj;
#if INLINE_ASM == 3

/*  esi = i
    edi = j
    ebx = data
    edx = temp
    ecx = prime
*/    

        ASM mov   ecx,DWORD PTR prime
        ASM mov   ebx,DWORD PTR data
        ASM mov   esi,DWORD PTR m
    hop1:
        ASM cmp   esi,DWORD PTR newn
        ASM jge   hop0
        ASM mov   edi,esi
        ASM add   edi,DWORD PTR mmax

        ASM mov   edx,[ebx+4*esi]
        ASM add   edx,ecx
        ASM sub   edx,[ebx+4*edi]
        ASM mov   eax,[ebx+4*esi]
        ASM add   eax,[ebx+4*edi]
        ASM cmp   eax,ecx
        ASM jl    hop2
        ASM sub   eax,ecx
    hop2:
        ASM mov   [ebx+4*esi],eax 

        ASM mov   eax,DWORD PTR w      
        ASM mul   edx                  
        ASM div   ecx
        ASM mov   [ebx+4*edi],edx 

        ASM add   esi,DWORD PTR istep
        ASM jmp   hop1
    hop0:
        ASM nop
#else
            for (i=m;i<newn;i+=istep) {
                j=i+mmax;
                temp=prime+data[i]-data[j];
                data[i]+=data[j];
                if (data[i]>=prime) data[i]-=prime;

#ifdef MR_NOASM
                dble=(mr_large)w*temp;
                data[j]=(mr_utype)(dble-(mr_large)prime*(dble/prime));
#else
                muldiv(w,temp,(mr_small)0,prime,(mr_small *)&data[j]);
#endif
            }
#endif
        }
    }
}

void mr_dit_fft(int logn,int pr,mr_utype *data)
{ /* decimate-in-time inverse fourier transform */
    int mmax,m,j,k,i,istep,ii,jj,newn,offset;
    mr_utype w,temp,prime,*roots;
#ifdef MR_NOASM
    mr_large dble;
#endif
    prime=mr_mip->prime[pr];
    roots=mr_mip->roots[pr];
    offset=(mr_mip->logN-logn);

    newn=(1<<logn);
    mmax=1;
    for (k=0;k<logn;k++) {
        istep=(mmax<<1); 
        ii=0;
        jj=newn/istep;
        ii+=jj;
        for (i=0;i<newn;i+=istep)
        { /* special case root=1 */
            j=i+mmax;
            temp=data[j];
            data[j]=data[i]-temp;
            if (data[j]<0) data[j]+=prime;
            data[i]+=temp;
            if (data[i]>prime) data[i]-=prime;
        }
        for (m=1;m<mmax;m++) {
            w=roots[(ii<<offset)-1];
            ii+=jj;
#if INLINE_ASM == 3

/*  esi = i
    edi = j
    ebx = data
    edx = temp
    ecx = prime
*/    


        ASM mov   ebx,DWORD PTR data
        ASM mov   ecx,DWORD PTR prime
        ASM mov   esi,DWORD PTR m
hop4:
        ASM cmp   esi,DWORD PTR newn
        ASM jge   hop5
        ASM mov   edi,esi
        ASM add   edi,DWORD PTR mmax

        ASM mov   eax,[ebx+4*edi]
        ASM mul   DWORD PTR w
        ASM div   ecx

        ASM cmp   [ebx+4*esi],ecx
        ASM jl    hop3
        ASM sub   [ebx+4*esi],ecx
hop3:   
        ASM mov   eax,[ebx+4*esi]
        ASM add   [ebx+4*esi],edx
        ASM add   eax,ecx
        ASM sub   eax,edx
        ASM mov   [ebx+4*edi],eax

        ASM add   esi,DWORD PTR istep
        ASM jmp   hop4
hop5:
        ASM nop
#else
            for (i=m;i<newn;i+=istep) {
                j=i+mmax;
#ifdef MR_NOASM
                dble=(mr_large)w*data[j];
                temp=(mr_utype)(dble-(mr_large)prime*(dble/prime));
#else
                muldiv(w,data[j],(mr_small)0,prime,(mr_small *)&temp);
#endif
                if (data[i]>=prime) data[i]-=prime;
                data[j]=prime+data[i]-temp;
                data[i] += temp;
            }
#endif
        }
        mmax=istep;
    }
}

static void modxn_1(int n,int deg,big *x)
{  /* set X (of degree deg) =X mod x^n-1 = X%x^n + X/x^n */
    int i;
    for (i=0;n+i<=deg;i++)
    {
        nres_modadd(x[i],x[n+i],x[i]);
        zero(x[n+i]);
    }
}

BOOL mr_poly_rem(int dg,big *G,big *R)
{ /* G is a polynomial of degree dg */
    int i,j,newn,logn,np,n;
    mr_utype p,inv,fac;
    big *q;
    n=mr_mip->degree;  /* degree of modulus */
    if (n==0) return FALSE; /* the preset tables have been destroyed */
    np=mr_mip->nprimes;

    newn=1; logn=0;
    while (2*n>newn) { newn<<=1; logn++; }

    q=(big *)mr_alloc(dg+1,sizeof(big));
    for (i=0;i<=dg;i++) q[i]=mirvar(0);

    for (i=0;i<np;i++)
    {
        p=mr_mip->prime[i];
        for (j=n;j<=dg;j++)
        {
            if (G[j]==NULL) mr_mip->t[i][j-n]=0;
            else mr_mip->t[i][j-n]=mr_sdiv(G[j],p,mr_mip->w1); 
        }
        for (j=dg-n+1;j<newn;j++) mr_mip->t[i][j]=0;
        mr_dif_fft(logn,i,mr_mip->t[i]);
        for (j=0;j<newn;j++)
        {
            muldiv(mr_mip->t[i][j],mr_mip->s1[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
        }
        mr_dit_fft(logn,i,mr_mip->t[i]);
        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn)
        { /* adjust 1/N log p for N/2, N/4 etc */
            fac=(1<<(mr_mip->logN-logn));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<n;j++)
            muldiv(mr_mip->t[i][j+n-1],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j+n-1]);
    }

    mr_shift(mr_mip->modulus,(int)mr_mip->modulus[0],mr_mip->w6);
       /* w6 = N.R */
    for (j=0;j<n;j++)
    {
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j+n-1];
        scrt(&mr_mip->chin,mr_mip->cr,q[j]);
        divide(q[j],mr_mip->w6,mr_mip->w6);  /* z[j] may be too big for redc */ 
        redc(q[j],q[j]);
    }

    for (i=0;i<np;i++)
    {
        p=mr_mip->prime[i];
        for (j=0;j<n;j++)
            mr_mip->t[i][j]=mr_sdiv(q[j],p,mr_mip->w1);
        for (j=n;j<1+newn/2;j++) mr_mip->t[i][j]=0;

        mr_dif_fft(logn-1,i,mr_mip->t[i]);  /* Note: Half size */
        for (j=0;j<newn/2;j++)
            muldiv(mr_mip->t[i][j],mr_mip->s2[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);

        mr_dit_fft(logn-1,i,mr_mip->t[i]);

        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn-1)
        {
            fac=(1<<(mr_mip->logN-logn+1));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<n;j++)
            muldiv(mr_mip->t[i][j],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);

    }

    for (i=0;i<=dg;i++) 
    { 
        if (G[i]!=NULL) copy(G[i],q[i]); 
        else            zero(q[i]); 
    } 

    modxn_1(newn/2,dg,q);    /* q=G mod 2^x - 1 */

    mr_shift(mr_mip->modulus,(int)mr_mip->modulus[0],mr_mip->w6);
       /* w6 = N.R */
    for (j=0;j<n;j++)
    {
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j];
        scrt(&mr_mip->chin,mr_mip->cr,R[j]);
        divide(R[j],mr_mip->w6,mr_mip->w6);  /* z[j] may be too big for redc */ 
        redc(R[j],R[j]);
        nres_modsub(q[j],R[j],R[j]);
    }

    for (i=0;i<=dg;i++) mr_free(q[i]);
    mr_free(q);

    return TRUE;
}

void mr_polymod_set(int n, big *rf,big *f)
{ /* n is degree of f */
    int i,j,np,newn,logn,degree;
    mr_utype p,inv,fac;
    big *F;
    degree=2*n;
    newn=1; logn=0;
    while (degree>newn) { newn<<=1; logn++; }

    if (mr_mip->degree!=0) 
    {
        for (i=0;i<mr_mip->nprimes;i++)
        {
            mr_free(mr_mip->s1[i]);
            mr_free(mr_mip->s2[i]);
        }
        mr_free(mr_mip->s1);
        mr_free(mr_mip->s2);
    }

    if (mr_mip->logN<logn)
        np=mr_fft_init(logn,mr_mip->modulus,mr_mip->modulus,TRUE);
    else np=mr_mip->nprimes;

    mr_mip->degree=n;
    mr_mip->s1=(mr_utype **)mr_alloc(np,sizeof(mr_utype *));
    mr_mip->s2=(mr_utype **)mr_alloc(np,sizeof(mr_utype *));
    
    F=(big *)mr_alloc(n+1,sizeof(big));
    for (i=0;i<=n;i++) 
    {
        F[i]=mirvar(0);
        if (f[i]!=NULL) copy(f[i],F[i]);
    }

    modxn_1(newn/2,n,F);

    for (i=0;i<np;i++)
    {
      /* reserve space for precomputed tables.   */
      /* Note that s2 is half the size of s1     */
        mr_mip->s1[i]=(mr_utype *)mr_alloc(newn,sizeof(mr_utype));
        mr_mip->s2[i]=(mr_utype *)mr_alloc(1+newn/2,sizeof(mr_utype));

        p=mr_mip->prime[i];
        for (j=0;j<n;j++)
        {
            if (rf[j]==NULL) mr_mip->s1[i][j]=0;
            else mr_mip->s1[i][j]=mr_sdiv(rf[j],p,mr_mip->w1);
        }
        mr_dif_fft(logn,i,mr_mip->s1[i]);
        
       for (j=0;j<=n;j++)
           mr_mip->s2[i][j]=mr_sdiv(F[j],p,mr_mip->w1);
       mr_dif_fft(logn-1,i,mr_mip->s2[i]);
    }
    for (i=0;i<=n;i++) mr_free(F[i]);
    mr_free(F);
}

int mr_ps_zzn_mul(int deg,big *x,big *y,big *z)
{
    int i,j,newn,logn,np,len;
    mr_utype inv,p,fac;
    newn=1; logn=0;
    while (2*deg>newn) { newn <<=1; logn++; }
    if (mr_mip->logN<logn)
        np=mr_fft_init(logn,mr_mip->modulus,mr_mip->modulus,TRUE);
    else np=mr_mip->nprimes;

    for (i=0;i<np;i++)
    {
        p=mr_mip->prime[i];
        for (j=0;j<deg;j++)
        {
            if (x[j]==NULL) mr_mip->w[j]=0;
            else mr_mip->w[j]=mr_sdiv(x[j],p,mr_mip->w1);
        }
        for (j=deg;j<newn;j++) mr_mip->w[j]=0;

        mr_dif_fft(logn,i,mr_mip->w);  
        for (j=0;j<deg;j++)
        {
            if (y[j]==NULL) mr_mip->t[i][j]=0;
            else mr_mip->t[i][j]=mr_sdiv(y[j],p,mr_mip->w1);
        }  
        for (j=deg;j<newn;j++) mr_mip->t[i][j]=0;
        mr_dif_fft(logn,i,mr_mip->t[i]);
  /* multiply FFTs */
        for (j=0;j<newn;j++)
            muldiv(mr_mip->w[j],mr_mip->t[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
        mr_dit_fft(logn,i,mr_mip->t[i]);           /* np*N*lgN */

        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn)
        {
            fac=(1<<(mr_mip->logN-logn));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<deg;j++)                   /* np*N */
            muldiv(mr_mip->t[i][j],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
    }
    mr_shift(mr_mip->modulus,(int)mr_mip->modulus[0],mr_mip->w6);
    for (j=0;j<deg;j++)
    {
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j];
        scrt(&mr_mip->chin,mr_mip->cr,z[j]);
        divide(z[j],mr_mip->w6,mr_mip->w6);
        redc(z[j],z[j]);
    }
    return np;
}

int mr_ps_big_mul(int deg,big *x,big *y,big *z)
{ /* Multiply two power series with large integer parameters */
    int i,j,newn,logn,np,len;
    mr_utype inv,p,fac;
    newn=1; logn=0;
    while (2*deg>newn) { newn <<=1; logn++; }

    zero(mr_mip->w2);
    zero(mr_mip->w4);

/* find biggest element in each series */
    for (i=0;i<deg;i++)
    {
        if (x[i]!=NULL) 
        {
            absol(x[i],mr_mip->w3);
            if (compare(mr_mip->w3,mr_mip->w2)>0) copy(mr_mip->w3,mr_mip->w2);
        }
        if (y[i]!=NULL)
        {
            absol(y[i],mr_mip->w3);
            if (compare(mr_mip->w3,mr_mip->w4)>0) copy(mr_mip->w3,mr_mip->w4);
        }
    }
    premult(mr_mip->w4,2,mr_mip->w4);     /* range is +ve and -ve */
                                          /* so extra factor of 2 included */

    np=mr_fft_init(logn,mr_mip->w4,mr_mip->w2,TRUE); 
    convert(1,mr_mip->w3);
/* compute coefficients modulo fft primes */    
    for (i=0;i<np;i++)
    {
        p=mr_mip->prime[i];
        mr_pmul(mr_mip->w3,p,mr_mip->w3); 
        for (j=0;j<deg;j++)
        {
            if (x[j]==NULL) mr_mip->w[j]=0;
            else 
            {
                if (size(x[j])>=0)
                {
                    copy(x[j],mr_mip->w1);
                    mr_mip->w[j]=mr_sdiv(mr_mip->w1,p,mr_mip->w1);
                }
                else
                {
                    negate(x[j],mr_mip->w1);
                    mr_mip->w[j]=p-mr_sdiv(mr_mip->w1,p,mr_mip->w1);
                }
            }
        }
        for (j=deg;j<newn;j++) mr_mip->w[j]=0;

        mr_dif_fft(logn,i,mr_mip->w);  
        for (j=0;j<deg;j++)
        {
            if (y[j]==NULL) mr_mip->t[i][j]=0;
            else 
            {
                if (size(y[j])>=0)
                {
                    copy(y[j],mr_mip->w1);
                    mr_mip->t[i][j]=mr_sdiv(mr_mip->w1,p,mr_mip->w1);
                }
                else
                {
                    negate(y[j],mr_mip->w1);
                    mr_mip->t[i][j]=p-mr_sdiv(mr_mip->w1,p,mr_mip->w1);
                }
            }
        }  
        for (j=deg;j<newn;j++) mr_mip->t[i][j]=0;
        mr_dif_fft(logn,i,mr_mip->t[i]);
  /* multiply FFTs */
        for (j=0;j<newn;j++)
            muldiv(mr_mip->w[j],mr_mip->t[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
        mr_dit_fft(logn,i,mr_mip->t[i]);           /* np*N*lgN */

        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn)
        {
            fac=(1<<(mr_mip->logN-logn));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<deg;j++)                   /* np*N */
            muldiv(mr_mip->t[i][j],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
    }
  /* w3 is product of chinese primes */
    decr(mr_mip->w3,1,mr_mip->w4);
    subdiv(mr_mip->w4,2,mr_mip->w4); /* find mid-point of range */
  
    for (j=0;j<deg;j++)
    {
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j];
        scrt(&mr_mip->chin,mr_mip->cr,z[j]); /* N*3*np*np/2  */
        if (compare(z[j],mr_mip->w4)>=0)
        { /* In higher half of range, so number is negative */
            subtract(mr_mip->w3,z[j],z[j]);
            negate(z[j],z[j]);
        }
    }            /* np*np*N/4 */
    return np;
}

int mr_poly_mul(int degx,big *x,int degy,big *y,big *z)
{ /*  Multiply two polynomials. The big arrays are of size degree */
    int i,j,newn,logn,np,len,degree;
    mr_utype inv,p,fac;
    degree=degx+degy;
    if (x==y) 
    {
        mr_poly_sqr(degx,x,z);
        return degree;
    }
    newn=1; logn=0;
    while (degree+1>newn) { newn<<=1; logn++; }

    if (mr_mip->logN<logn)
        np=mr_fft_init(logn,mr_mip->modulus,mr_mip->modulus,TRUE);
    else np=mr_mip->nprimes;

/* compute coefficients modulo fft primes */    
    for (i=0;i<np;i++)
    {                                              
        p=mr_mip->prime[i];
        for (j=0;j<=degx;j++)
        {
            if (x[j]==NULL) mr_mip->w[j]=0;
            else mr_mip->w[j]=mr_sdiv(x[j],p,mr_mip->w1);      /* np*np*N/2 muldivs */
        }
        for (j=degx+1;j<newn;j++) mr_mip->w[j]=0;
        mr_dif_fft(logn,i,mr_mip->w);              /* np*N*lgN     */

        for (j=0;j<=degy;j++)
        {
            if (y[j]==NULL) mr_mip->t[i][j]=0;
            else mr_mip->t[i][j]=mr_sdiv(y[j],p,mr_mip->w1);   /* np*np*N/2   */
        }
        for (j=degy+1;j<newn;j++) mr_mip->t[i][j]=0;
        mr_dif_fft(logn,i,mr_mip->t[i]);           /* np*N*lgN */

   /* multiply FFTs */
        for (j=0;j<newn;j++)                       /* np*N */
            muldiv(mr_mip->w[j],mr_mip->t[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
        mr_dit_fft(logn,i,mr_mip->t[i]);           /* np*N*lgN */

        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn)
        {
            fac=(1<<(mr_mip->logN-logn));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<=degree;j++)                   /* np*N */
            muldiv(mr_mip->t[i][j],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
    }
    mr_shift(mr_mip->modulus,(int)mr_mip->modulus[0],mr_mip->w6);
       /* w6 = N.R */
    for (j=0;j<=degree;j++)
    {
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j];
        scrt(&mr_mip->chin,mr_mip->cr,z[j]); /* N*3*np*np/2 */
        divide(z[j],mr_mip->w6,mr_mip->w6);  /* z[j] may be too big for redc */ 
        redc(z[j],z[j]);
    }                                        /* np*np*N/4 */

    return degree;
}

int mr_poly_sqr(int degx,big *x,big *z)
{ /*  Multiply two polynomials. The big arrays are of size degree */
    int i,j,newn,logn,np,len,degree;
    mr_utype inv,p,fac;
    degree=2*degx;
    newn=1; logn=0;
    while (degree+1>newn) { newn<<=1; logn++; }

    if (mr_mip->logN<logn)
        np=mr_fft_init(logn,mr_mip->modulus,mr_mip->modulus,TRUE);
    else np=mr_mip->nprimes;

/* compute coefficients modulo fft primes */    
    for (i=0;i<np;i++)
    {
        p=mr_mip->prime[i];
        for (j=0;j<=degx;j++)
        {
            if (x[j]==NULL) mr_mip->t[i][j]=0;
            else mr_mip->t[i][j]=mr_sdiv(x[j],p,mr_mip->w1);
        }
        for (j=degx+1;j<newn;j++) mr_mip->t[i][j]=0;
        mr_dif_fft(logn,i,mr_mip->t[i]);

   /* multiply FFTs */
        for (j=0;j<newn;j++)
            muldiv(mr_mip->t[i][j],mr_mip->t[i][j],(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
        mr_dit_fft(logn,i,mr_mip->t[i]);

        inv=mr_mip->inverse[i];
        if (mr_mip->logN > logn)
        { /* adjust 1/N log p for smaller N  */
            fac=(1<<(mr_mip->logN-logn));
            inv=smul(fac,inv,p);
        }
        for (j=0;j<=degree;j++)
            muldiv(mr_mip->t[i][j],inv,(mr_small)0,p,(mr_small *)&mr_mip->t[i][j]);
    }
    mr_shift(mr_mip->modulus,(int)mr_mip->modulus[0],mr_mip->w6);
       /* w6 = N.R */
    for (j=0;j<=degree;j++)
    { /* apply CRT to each column */
        for (i=0;i<np;i++)
            mr_mip->cr[i]=mr_mip->t[i][j];

        scrt(&mr_mip->chin,mr_mip->cr,z[j]);
        divide(z[j],mr_mip->w6,mr_mip->w6);  /* z[j] may be too big for redc */ 
        redc(z[j],z[j]);
    }

    return degree;
}

static BOOL init_it(int logn)
{ /* find primes, table of roots, inverses etc for new N */
    mr_small x[3];
   
    x[0]=2; x[1]=0; x[2]=1; 
    if (mr_fft_init(logn,x,x,FALSE)!=3) return FALSE;

    mr_mip->const1=invers(mr_mip->prime[0],mr_mip->prime[1]);
    mr_mip->const2=invers(mr_mip->prime[0],mr_mip->prime[2]);
    mr_mip->const3=invers(mr_mip->prime[1],mr_mip->prime[2]);
    if (mr_mip->base==0) mr_mip->msw=muldvd(mr_mip->prime[0],mr_mip->prime[1],(mr_small)0,&mr_mip->lsw);
    else                 mr_mip->msw=muldiv(mr_mip->prime[0],mr_mip->prime[1],(mr_small)0,mr_mip->base,&mr_mip->lsw);
    mr_mip->logN=logn;
    return TRUE;
}


void fft_mult(big x,big y,big z)
{ /* "fast" O(n.log n) multiplication */
    int i,pr,xl,yl,zl,newn,logn;
    mr_small v1,v2,v3,ic,c1,c2,sz,p,fac,inv;
    mr_utype *work,*d[3],*dptr,*d0,*d1,*d2,t;
    if (mr_mip->ERNUM) return;
    if (y[0]==0 || x[0]==0) 
    {
        zero(z);
        return;
    }

    MR_IN(72)

    if (mr_notint(x) || mr_notint(y))
    {
        mr_berror(MR_ERR_INT_OP);
        MR_OUT
        return;
    }  
    sz=((x[0]&mr_mip->MSBIT)^(y[0]&mr_mip->MSBIT));
    xl=(int)(x[0]&mr_mip->OBITS);
    yl=(int)(y[0]&mr_mip->OBITS);
    zl=xl+yl;
    if (xl<64 || yl<64)
    { /* not worth it! */
        multiply(x,y,z);
        MR_OUT
        return;
    }
    if (zl>mr_mip->nib && mr_mip->check)
    {
        mr_berror(MR_ERR_OVERFLOW);
        MR_OUT
        return;
    }
    newn=1; logn=0;
    while (zl>newn) { newn<<=1; logn++;}
    if (logn>mr_mip->logN)     /*  2^(N+1) settings can be used for 2^N */
    { /* numbers too big for current settings */
        if (!init_it(logn)) 
        {
            mr_berror(MR_ERR_OUT_OF_MEMORY);
            MR_OUT
            return;
        }
    }
    if (newn>2*mr_mip->nib)
    {
        mr_berror(MR_ERR_OVERFLOW);
        MR_OUT
        return;
    }

    d0=mr_mip->t[0]; d1=mr_mip->t[1]; d2=mr_mip->t[2];
    work=mr_mip->w;

    fac=(1<<(mr_mip->logN-logn));
    
    for (pr=0;pr<3;pr++)
    { /* multiply mod each prime */
        p=mr_mip->prime[pr];
        inv=mr_mip->inverse[pr];  
        if (fac!=1) inv=smul(fac,inv,p);  /* adjust 1/N mod p */
        
        dptr=mr_mip->t[pr];
        for (i=1;i<=xl;i++) dptr[i-1]=x[i]%p;
        for (i=xl;i<newn;i++) dptr[i]=0;

        mr_dif_fft(logn,pr,dptr);
        if (x!=y)
        {
            for (i=1;i<=yl;i++) work[i-1]=y[i]%p;
            for (i=yl;i<newn;i++) work[i]=0;
            mr_dif_fft(logn,pr,work);
        }
        else work=dptr;
        for (i=0;i<newn;i++)
        {  /* "multiply" Fourier transforms */
            muldiv(dptr[i],work[i],(mr_small)0,p,(mr_small *)&dptr[i]);
        }
        mr_dit_fft(logn,pr,dptr);

        for (i=0;i<newn;i++)
        {  /* convert to mixed radix form     *
            * using chinese remainder thereom *
            * but first multiply by 1/N mod p */
            muldiv(dptr[i],inv,(mr_small)0,p,(mr_small *)&dptr[i]); 
            if (pr==1)
            {
                t=d1[i]-d0[i];
                while (t<0) t+=mr_mip->prime[1];
                muldiv(t,mr_mip->const1,(mr_small)0,mr_mip->prime[1],(mr_small *)&d1[i]);
            } 
            if (pr==2)
            {
                t=d2[i]-d0[i];
                while (t<0) t+=mr_mip->prime[2];
                muldiv(t,mr_mip->const2,(mr_small)0,mr_mip->prime[2],(mr_small *)&t);
                t-=d1[i];
                while (t<0) t+=mr_mip->prime[2];
                muldiv(t,mr_mip->const3,(mr_small)0,mr_mip->prime[2],(mr_small *)&d2[i]);
            }
        }    
    }

    zero(z);
    c1=c2=0;
    if (mr_mip->base==0) for (i=0;i<zl;i++)
    { /* propagate carries */
        v1=d0[i];
        v2=d1[i];
        v3=d2[i];
        v2=muldvd(v2,mr_mip->prime[0],v1,&v1);
        c1+=v1;
        if (c1<v1) v2++;
        ic=c2+muldvd(mr_mip->lsw,v3,c1,&z[i+1]);
        c2=muldvd(mr_mip->msw,v3,ic,&c1);
        c1+=v2;
        if (c1<v2) c2++;
    }
    else for (i=0;i<zl;i++)
    { /* propagate carries */
        v1=d0[i];
        v2=d1[i];
        v3=d2[i];
        v2=muldiv(v2,mr_mip->prime[0],v1+c1,mr_mip->base,&v1);
        ic=c2+muldiv(mr_mip->lsw,v3,v1,mr_mip->base,&z[i+1]);
        c2=muldiv(mr_mip->msw,v3,v2+ic,mr_mip->base,&c1);
    }
    z[0]=(sz|zl); /* set length and sign of result */
    mr_lzero(z);
    MR_OUT
}

/*
main()
{
    big x,y,z,w;
    int i,j,k;
    mirsys(500,0);
    x=mirvar(0);
    y=mirvar(0);
    z=mirvar(0);
    w=mirvar(0);

    bigdig(64*10,2,x);
    bigdig(64*10,2,y);
    multiply(x,y,z);
    cotnum(z,stdout);

    fft_mult(x,y,w);
    cotnum(w,stdout);
    if (compare(z,w)!=0) printf("Problems\n");



    bigdig(32*10,2,x);
    bigdig(32*10,2,y);
    multiply(x,y,z);
    cotnum(z,stdout);

    fft_mult(x,y,w);
    cotnum(w,stdout);
    if (compare(z,w)!=0) printf("Problems\n");

    bigdig(16*10,2,x);
    bigdig(16*10,2,y);
    multiply(x,y,z);
    cotnum(z,stdout);

    fft_mult(x,y,w);
    cotnum(w,stdout);
    if (compare(z,w)!=0) printf("Problems\n");
}
*/

