/*
 *   MIRACL methods for modular exponentiation
 *   mrpower.c 
 *
 *   Copyright (c) 1988-1999 Shamus Software Ltd.
 */

#include <stdio.h>
#include <miracl.h>

extern miracl *mr_mip;

static int mr_window(big x,int i,int *nbs,int * nzs)
{ /* returns sliding window value, max. of 5 bits,         *
   * starting at i-th bit of big x. nbs is number of bits  *
   * processed, nzs is the number of additional trailing   *
   * zeros detected. Returns valid bit pattern 1x..x1 with *
   * no two adjacent 0's. So 10101 will return 21 with     *
   * nbs=5, nzs=0. 11001 will return 3, with nbs=2, nzs=2, *
   * having stopped after the first 11..  */

    int j,r,w;
    w=5;

/* check for leading 0 bit */

    *nbs=1;
    *nzs=0;
    if (!mr_testbit(x,i)) return 0;

/* adjust window size if not enough bits left */
   
    if (i-w+1<0) w=i+1;

    r=1;
    for (j=i-1;j>i-w;j--)
    { /* accumulate bits. Abort if two 0's in a row */
        (*nbs)++;
        r*=2;
        if (mr_testbit(x,j)) r+=1;
        if (r%4==0)
        { /* oops - too many zeros - shorten window */
            r/=4;
            *nbs-=2;
            *nzs=2;
            break;
        }
    }
    if (r%2==0)
    { /* remove trailing 0 */
        r/=2;
        *nzs=1;
        (*nbs)--;
    }
    return r;
}

static int mr_window2(big x,big y,int i,int *nbs,int *nzs)
{ /* two bit window for double exponentiation */
    int r,w;
    BOOL a,b,c,d;
    w=2;
    *nbs=1;
    *nzs=0;

/* check for two leading 0's */
    a=mr_testbit(x,i); b=mr_testbit(y,i);

    if (!a && !b) return 0;
    if (i<1) w=1;

    if (a)
    {
        if (b) r=3;
        else   r=2;
    }
    else r=1;
    if (w==1) return r;

    c=mr_testbit(x,i-1); d=mr_testbit(y,i-1);

    if (!c && !d) 
    {
        *nzs=1;
        return r;
    }

    *nbs=2;
    r*=4;
    if (c)
    {
        if (d) r+=3;
        else   r+=2;
    }
    else r+=1;
    return r;
}

void nres_powltr(int x,big y,big w)
{ /* calculates w=x^y mod z using Left to Right Method   */
  /* uses only n^2 modular squarings, because x is small */
  /* Note: x is NOT an nresidue */
    int i,nb;

    if (mr_mip->ERNUM) return;
    copy(y,mr_mip->w1);

    MR_IN(86)

    zero(w);
    if (x==0) 
    {
        if (size(mr_mip->w1)==0) 
        { /* 0^0 = 1 */
            convert(1,w);
            nres(w,w);
        }
        MR_OUT
        return;
    }

    convert(1,w);
    nres(w,w);
    if (size(mr_mip->w1)==0) 
    {
        MR_OUT
        return;
    }
    if (size(mr_mip->w1)<0) mr_berror(MR_ERR_NEG_POWER);
    if (mr_mip->ERNUM)
    {
        MR_OUT
        return;
    }

    if (mr_mip->base==mr_mip->base2)
    { 
        nb=logb2(mr_mip->w1);
        convert(x,w);
        nres(w,w);
        if (nb>1) for (i=nb-2;i>=0;i--)
        { /* Left to Right binary method */

            if (mr_mip->user!=NULL) (*mr_mip->user)();

            nres_modmult(w,w,w);
            if (mr_testbit(mr_mip->w1,i))
            { /* this is quick */
                premult(w,x,w);
                divide(w,mr_mip->modulus,mr_mip->modulus);
            }
        }
    }    
    else
    {
        expint(2,logb2(mr_mip->w1)-1,mr_mip->w2);
        while (size(mr_mip->w2)!=0)
        { /* Left to Right binary method */

            if (mr_mip->user!=NULL) (*mr_mip->user)();

            nres_modmult(w,w,w);
            if (compare(mr_mip->w1,mr_mip->w2)>=0)
            {
                premult(w,x,w);
                divide(w,mr_mip->modulus,mr_mip->modulus);
                subtract(mr_mip->w1,mr_mip->w2,mr_mip->w1);
            }
            subdiv(mr_mip->w2,2,mr_mip->w2);
        }
    }
    MR_OUT
    return;
}

void nres_powmod(big x,big y,big w)
{  /*  calculates w=x^y mod z, using m-residues       *
    *  See "Analysis of Sliding Window Techniques for *
    *  Exponentiation, C.K. Koc, Computers. Math. &   *
    *  Applic. Vol. 30 pp17-24 1995. Uses work-space  *
    *  variables for pre-computed table. */
    int i,j,k,t,nb,nbw,nzs,n;
#ifdef MR_PENTIUM
    double *dtable[16],*dw,*dw2,*p1,*p2;
#endif
    big table[16];
    if (mr_mip->ERNUM) return;

    copy(y,mr_mip->w1);
    copy(x,mr_mip->w3);

    MR_IN(84)
    zero(w);
    if (size(x)==0)
    {
       if (size(mr_mip->w1)==0)
       { /* 0^0 = 1 */
           convert(1,w);
           nres(w,w);
       } 
       MR_OUT
       return;
    }

    convert(1,w);
    nres(w,w);
    if (size(mr_mip->w1)==0) 
    {
        MR_OUT
        return;
    }

    if (size(mr_mip->w1)<0) mr_berror(MR_ERR_NEG_POWER);

    if (mr_mip->ERNUM)
    {
        MR_OUT
        return;
    }
    if (mr_mip->base==mr_mip->base2)
    { /* build a table. Up to 5-bit sliding windows. Windows with
       * two adjacent 0 bits simply won't happen */
#ifdef MR_PENTIUM
      if (mr_mip->ACTIVE)
      {
        mr_set_align(sizeof(double));  /* set for correct alignment */

        dw=mr_alloc(mr_mip->nib+1,sizeof(double));
        
        dw2=mr_alloc(mr_mip->nib+1,sizeof(double));

        for (i=0;i<4;i++) dtable[i]=mr_alloc(mr_mip->nib+1,sizeof(double));
        dtable[4]=NULL;
        for (i=5;i<8;i++) dtable[i]=mr_alloc(mr_mip->nib+1,sizeof(double));
        dtable[8]=NULL;
        dtable[9]=NULL;
        for (i=10;i<12;i++) dtable[i]=mr_alloc(mr_mip->nib+1,sizeof(double));
        dtable[12]=NULL;
        for (i=13;i<16;i++) dtable[i]=mr_alloc(mr_mip->nib+1,sizeof(double));

        mr_mip->dw0=mr_alloc(2*mr_mip->nib+3,sizeof(double));
        mr_mip->dmodulus=mr_alloc(mr_mip->nib+1,sizeof(double));

        for (i=0;i<=mr_mip->modulus[0];i++) mr_mip->dmodulus[i]=(double)mr_mip->modulus[i];
        p1=dtable[0];
        for (i=0;i<=mr_mip->w3[0];i++) 
        {
            dw[i]=(double)mr_mip->w3[i];
            p1[i]=(double)mr_mip->w3[i];
        }
        fastmodsquare(dw,dw2);

        n=15;
        j=0;
        do
        { /* pre-computations */
            t=1; k=j+1;
            while (dtable[k]==NULL) {k++; t++;}

            p1=dtable[j]; p2=dtable[k];
            for (i=0;i<=p1[0];i++) p2[i]=p1[i];

            for (i=0;i<t;i++) fastmodmult(dtable[k],dw2,dtable[k]);
            j=k;
        } while (j<n);

        nb=logb2(mr_mip->w1);
        if (nb>1) for (i=nb-2;i>=0;)
        { /* Left to Right method */
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            n=mr_window(mr_mip->w1,i,&nbw,&nzs); 
            for (j=0;j<nbw;j++)
                    fastmodsquare(dw,dw);
            if (n>0) fastmodmult(dw,dtable[n/2],dw); 
            i-=nbw;
            if (nzs)
            {
                for (j=0;j<nzs;j++) fastmodsquare(dw,dw);
                i-=nzs;
            }
        }
        for (i=0;i<=(int)dw[0];i++) w[i]=(int)dw[i];

        mr_free(dw);
        mr_free(dw2);
        mr_free(mr_mip->dmodulus);
        mr_free(mr_mip->dw0);
        for (i=0;i<16;i++) if (dtable[i]!=NULL) mr_free(dtable[i]);

        mr_set_align(0);
      }
      else
#endif
      {
        zero(mr_mip->w5);   /* break double-length ......*/
        zero(mr_mip->w6);   /* work-space variables into 2 halves      */ 
        table[0]=mr_mip->w3; table[1]=mr_mip->w4; table[2]=mr_mip->w5; table[3]=mr_mip->w5d;
        table[4]=NULL;  table[5]=mr_mip->w6; table[6]=mr_mip->w6d; table[7]=mr_mip->w8;
        table[8]=NULL;  table[9]=NULL;  table[10]=mr_mip->w9; table[11]=mr_mip->w10;
        table[12]=NULL; table[13]=mr_mip->w11; table[14]=mr_mip->w12; table[15]=mr_mip->w13;

        nres_modmult(mr_mip->w3,mr_mip->w3,mr_mip->w2);  /* x^2 */
        n=15;
        j=0;
        do
        { /* pre-computations */
            t=1; k=j+1;
            while (table[k]==NULL) {k++; t++;}
            copy(table[j],table[k]);
            for (i=0;i<t;i++) nres_modmult(table[k],mr_mip->w2,table[k]);
            j=k;
        } while (j<n);

        nb=logb2(mr_mip->w1);
        copy(mr_mip->w3,w);
        if (nb>1) for (i=nb-2;i>=0;)
        { /* Left to Right method */

            if (mr_mip->user!=NULL) (*mr_mip->user)();

            n=mr_window(mr_mip->w1,i,&nbw,&nzs); 
            for (j=0;j<nbw;j++)
                    nres_modmult(w,w,w);
            if (n>0) nres_modmult(w,table[n/2],w); 
            i-=nbw;
            if (nzs)
            {
                for (j=0;j<nzs;j++) nres_modmult(w,w,w);
                i-=nzs;
            }
        }
        zero(mr_mip->w5d);
        zero(mr_mip->w6d);
      }
    }
    else
    {
        copy(mr_mip->w3,mr_mip->w2);
        forever
        { /* "Russian peasant" Right-to-Left exponentiation */

            if (mr_mip->user!=NULL) (*mr_mip->user)();

            if (subdiv(mr_mip->w1,2,mr_mip->w1)!=0)
                nres_modmult(w,mr_mip->w2,w);
            if (mr_mip->ERNUM || size(mr_mip->w1)==0) break;
            nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w2);
        }
    }
    MR_OUT
}

void nres_powmodn(int n,big *x,big *y,big w)
{
    int i,j,k,m,nb,ea;
    big *G;
    if (mr_mip->ERNUM) return;

    MR_IN(112)

    m=1<<n;
    G=(big *)mr_alloc(m,sizeof(big));

/* 2^n - n - 1 modmults */
/* 4(n=3) 11(n=4) etc   */

    for (i=0,k=1;i<n;i++)
    {
        for (j=0; j < (1<<i) ;j++)
        {
            G[k]=mirvar(0);
            if (j==0) copy(x[i],G[k]);
            else      nres_modmult(G[j],x[i],G[k]);
            k++;
        }
    }

    nb=0;
    for (j=0;j<n;j++) 
        if ((k=logb2(y[j]))>nb) nb=k;

    convert(1,w);
    nres(w,w);

    if (mr_mip->base==mr_mip->base2)
    {

        for (i=nb-1;i>=0;i--) 
        {
            if (mr_mip->user!=NULL) (*mr_mip->user)();
            ea=0;
            k=1;
            for (j=0;j<n;j++)
            {
                if (mr_testbit(y[j],i)) ea+=k;
                k<<=1;
            }
            nres_modmult(w,w,w);
            if (ea!=0) nres_modmult(w,G[ea],w);
        }
    }
    else mr_berror(MR_ERR_NOT_SUPPORTED);
    for (i=1;i<m;i++) mirkill(G[i]);
    mr_free(G);

    MR_OUT
}

void powmodn(int n,big *x,big *y,big p,big w)
{/* w=x[0]^y[0].x[1]^y[1] .... x[n-1]^y[n-1] mod n */
    int j;
    if (mr_mip->ERNUM) return;

    MR_IN(113)

    prepare_monty(p);
    for (j=0;j<n;j++) nres(x[j],x[j]);
    nres_powmodn(n,x,y,w);   
    for (j=0;j<n;j++) redc(x[j],x[j]);

    redc(w,w);

    MR_OUT
}

void nres_powmod2(big x,big y,big a,big b,big w)
{ /* finds w = x^y.a^b mod n. Fast for some cryptosystems */ 
    int i,j,nb,nb2,nbw,nzs,n;
    big table[16];
    if (mr_mip->ERNUM) return;
    copy(y,mr_mip->w1);
    copy(x,mr_mip->w2);
    copy(b,mr_mip->w3);
    copy(a,mr_mip->w4);
    zero(w);
    if (size(mr_mip->w2)==0 || size(mr_mip->w4)==0) return;

    MR_IN(99)

    convert(1,w);
    nres(w,w);
    if (size(mr_mip->w1)==0 && size(mr_mip->w3)==0) 
    {
        MR_OUT
        return;
    }
    if (size(mr_mip->w1)<0 || size(mr_mip->w3)<0) mr_berror(MR_ERR_NEG_POWER);
    if (mr_mip->ERNUM)
    {
        MR_OUT
        return;
    }
     
    if (mr_mip->base==mr_mip->base2)
    { /* uses 2-bit sliding window. This is 25% faster! */

        zero(mr_mip->w5);   /* break double-length ......*/
        zero(mr_mip->w6);   /* work-space variables into 2 halves      */ 

        nres_modmult(mr_mip->w2,mr_mip->w4,mr_mip->w5);
        nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w12);
        nres_modmult(mr_mip->w4,mr_mip->w4,mr_mip->w13);
        nres_modmult(mr_mip->w4,mr_mip->w13,mr_mip->w5d);
        nres_modmult(mr_mip->w2,mr_mip->w13,mr_mip->w6);
        nres_modmult(mr_mip->w6,mr_mip->w4,mr_mip->w6d);
        nres_modmult(mr_mip->w4,mr_mip->w12,mr_mip->w8);
        nres_modmult(mr_mip->w2,mr_mip->w12,mr_mip->w9);
        nres_modmult(mr_mip->w4,mr_mip->w9,mr_mip->w10);
        nres_modmult(mr_mip->w5d,mr_mip->w12,mr_mip->w11);
        nres_modmult(mr_mip->w9,mr_mip->w13,mr_mip->w12);
        nres_modmult(mr_mip->w10,mr_mip->w13,mr_mip->w13);

        table[0]=NULL; table[1]=mr_mip->w4;  table[2]=mr_mip->w2;   table[3]=mr_mip->w5;
        table[4]=NULL; table[5]=mr_mip->w5d; table[6]=mr_mip->w6;   table[7]=mr_mip->w6d;
        table[8]=NULL; table[9]=mr_mip->w8;  table[10]=mr_mip->w9;  table[11]=mr_mip->w10;
        table[12]=NULL;table[13]=mr_mip->w11;table[14]=mr_mip->w12; table[15]=mr_mip->w13;
        nb=logb2(mr_mip->w1);
        if ((nb2=logb2(mr_mip->w3))>nb) nb=nb2;

        for (i=nb-1;i>=0;)
        { 
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            n=mr_window2(mr_mip->w1,mr_mip->w3,i,&nbw,&nzs);
            for (j=0;j<nbw;j++)
                nres_modmult(w,w,w);
            if (n>0) nres_modmult(w,table[n],w);
            i-=nbw;
            if (nzs) 
            {
                nres_modmult(w,w,w);
                i--;
            } 
        }
        zero(mr_mip->w5d);
        zero(mr_mip->w6d);

    }
    else
    {
        nres_modmult(mr_mip->w2,mr_mip->w4,mr_mip->w5);

        if (compare(mr_mip->w1,mr_mip->w3)>=0)
             expint(2,logb2(mr_mip->w1)-1,mr_mip->w6);
        else expint(2,logb2(mr_mip->w3)-1,mr_mip->w6);
        while (size(mr_mip->w6)!=0)
        {
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            nres_modmult(w,w,w);
            if (compare(mr_mip->w1,mr_mip->w6)>=0)
            {
                if (compare(mr_mip->w3,mr_mip->w6)>=0)
                {
                     nres_modmult(w,mr_mip->w5,w);
                     subtract(mr_mip->w3,mr_mip->w6,mr_mip->w3);
                }

                else nres_modmult(w,mr_mip->w2,w);
                subtract(mr_mip->w1,mr_mip->w6,mr_mip->w1);
            }
            else
            {
                if (compare(mr_mip->w3,mr_mip->w6)>=0)
                {
                     nres_modmult(w,mr_mip->w4,w);
                     subtract(mr_mip->w3,mr_mip->w6,mr_mip->w3);
                }
            }
            subdiv(mr_mip->w6,2,mr_mip->w6);
        }
    }
    MR_OUT
}

void powmod2(big x,big y,big a,big b,big n,big w)
{/* w=x^y.a^b mod n */
    if (mr_mip->ERNUM) return;

    MR_IN(79)

    prepare_monty(n);
    nres(x,mr_mip->w2);
    nres(a,mr_mip->w4);
    nres_powmod2(mr_mip->w2,y,mr_mip->w4,b,w);
    redc(w,w);

    MR_OUT   
}

static mr_small sgcd(mr_small x,mr_small y)
{ /* integer GCD, returns GCD of x and y */
    mr_small r;
    if (y==(mr_small)0) return x;
    while ((r=x%y)!=(mr_small)0)
        x=y,y=r;
    return y;
}


void powmod(big x,big y,big n,big w)
{ /* fast powmod, using Montgomery's method internally */

    mr_small norm;
    BOOL mty;
    if (mr_mip->ERNUM) return;

    MR_IN(18)

    mty=TRUE;
    if (mr_mip->base!=mr_mip->base2)
    {
        if (size(n)<2 || sgcd(n[1],mr_mip->base)!=1) mty=FALSE;
    }
    else
        if (subdivisible(n,2)) mty=FALSE;


    if (!mty)
    { /* can't use Montgomery */
        copy(y,mr_mip->w1);
        copy(x,mr_mip->w3);
        zero(w);
        if (size(mr_mip->w3)==0) 
        {
            MR_OUT
            return;
        }
        convert(1,w);
        if (size(mr_mip->w1)==0) 
        {
            MR_OUT
            return;
        }
        if (size(mr_mip->w1)<0) mr_berror(MR_ERR_NEG_POWER);
        if (w==n)           mr_berror(MR_ERR_BAD_PARAMETERS) ;
        if (mr_mip->ERNUM)
        {
            MR_OUT
            return;
        }

        norm=normalise(n,n);
        divide(mr_mip->w3,n,n);
        forever
        { 
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            if (subdiv(mr_mip->w1,2,mr_mip->w1)!=0)
                mad(w,mr_mip->w3,mr_mip->w3,n,n,w);
            if (mr_mip->ERNUM || size(mr_mip->w1)==0) break;
            mad(mr_mip->w3,mr_mip->w3,mr_mip->w3,n,n,mr_mip->w3);
        }
        if (norm!=1)
        {
            mr_sdiv(n,norm,n);
            divide(w,n,n);
        }
    }
    else
    { /* optimized code for odd moduli */
        prepare_monty(n); 
        nres(x,mr_mip->w3);
        nres_powmod(mr_mip->w3,y,w);
        redc(w,w);
    }
    
    MR_OUT
}

int powltr(int x, big y, big n, big w)
{
    mr_small norm;
    BOOL mty;
    if (mr_mip->ERNUM) return 0;

    MR_IN(71)

    mty=TRUE;
    if (mr_mip->base!=mr_mip->base2)
    {
        if (size(n)<2 || sgcd(n[1],mr_mip->base)!=1) mty=FALSE;
    }
    else
    {
        if (subdivisible(n,2)) mty=FALSE;
    }

/* Montgomery modulus is in use... */
    if (mr_mip->modulus!=NULL)
        if (compare(n,mr_mip->modulus)!=0) mty=FALSE;

    if (!mty)
    { /* can't use Montgomery! */
        copy(y,mr_mip->w1);
        zero(w);
        if (x==0) 
        {
            MR_OUT
            return 0;
        }
        convert(1,w);
        if (size(mr_mip->w1)==0)
        {
            MR_OUT
            return 1;
        }
        
        if (size(mr_mip->w1)<0) mr_berror(MR_ERR_NEG_POWER);
        if (w==n)               mr_berror(MR_ERR_BAD_PARAMETERS) ;
        if (mr_mip->ERNUM)
        {
            MR_OUT
            return 0;
        }
        
        norm=normalise(n,n);

        expint(2,logb2(mr_mip->w1)-1,mr_mip->w2);
        while (size(mr_mip->w2)!=0)
        { /* Left to Right binary method */

            if (mr_mip->user!=NULL) (*mr_mip->user)();
            mad(w,w,w,n,n,w);
            if (compare(mr_mip->w1,mr_mip->w2)>=0)
            {
                premult(w,x,w);
                divide(w,n,n);
                subtract(mr_mip->w1,mr_mip->w2,mr_mip->w1);
            }
            subdiv(mr_mip->w2,2,mr_mip->w2);
        }
        if (norm!=1) 
        {
            mr_sdiv(n,norm,n);
            divide(w,n,n);
        }
    }
    else
    { /* optimized code for odd moduli */
        prepare_monty(n);
        nres_powltr(x,y,w);
        redc(w,w);
    }
    MR_OUT 
    return (size(w));
}

BOOL nres_sqroot(big x,big w)
{ /* w=sqrt(x) mod p. This depends on p being prime! */
    int i,n,e,r,cat;
    BOOL pp;
    copy(x,mr_mip->w15);
    zero(w);
    if (size(mr_mip->w15)==0) zero(w);
    if (mr_mip->ERNUM) return FALSE;

    MR_IN(100)

    convert(1,w);
    nres(w,w);
    if (compare(w,mr_mip->w15)==0)
    {
        MR_OUT
        return TRUE;
    }

    cat=remain(mr_mip->modulus,8);
    switch(cat)
    {
    case 0: case 2: case 4: case 6: 
        zero(w);
        MR_OUT
        return FALSE;

    case 3: case 7:                          /* easy case */
        incr(mr_mip->modulus,1,mr_mip->w14);
        subdiv(mr_mip->w14,4,mr_mip->w14);
        nres_powmod(x,mr_mip->w14,w);
        nres_modmult(w,w,mr_mip->w14);
        MR_OUT
        if (compare(mr_mip->w14,mr_mip->w15)==0) 
            return TRUE;
        zero(w);
        return FALSE;

    case 5:
        nres_modadd(mr_mip->w15,mr_mip->w15,mr_mip->w15); /* 2x */
        decr(mr_mip->modulus,5,mr_mip->w14);
        subdiv(mr_mip->w14,8,mr_mip->w14);
        nres_powmod(mr_mip->w15,mr_mip->w14,w);
        nres_modmult(w,w,mr_mip->w14);
        nres_modmult(mr_mip->w15,mr_mip->w14,mr_mip->w14);
        convert(1,mr_mip->w1);
        nres(mr_mip->w1,mr_mip->w1);
        nres_modsub(mr_mip->w14,mr_mip->w1,mr_mip->w14);
        nres_modmult(mr_mip->w14,w,w);
        if (!subdivisible(mr_mip->w15,2))
            add(mr_mip->w15,mr_mip->modulus,mr_mip->w15);
        subdiv(mr_mip->w15,2,mr_mip->w15);                /* x */
        nres_modmult(w,mr_mip->w15,w);
        nres_modmult(w,w,mr_mip->w14);
        MR_OUT
        if (compare(mr_mip->w14,mr_mip->w15)==0) 
            return TRUE;
        zero(w);
        return FALSE;
        
    case 1:                      /* difficult case. Shank's method */

        decr(mr_mip->modulus,1,mr_mip->w14);
        e=0;
        while (subdivisible(mr_mip->w14,2))
        {
            subdiv(mr_mip->w14,2,mr_mip->w14);
            e++;
        }
        for (r=2;;r++)
        {
            convert(1,mr_mip->w3);
            nres(mr_mip->w3,mr_mip->w3);             /* w3=1 */
            nres_powltr(r,mr_mip->w14,w);
            if (compare(w,mr_mip->w3)==0) continue;
            copy(w,mr_mip->w4);
            nres_negate(mr_mip->w3,mr_mip->w1);      /* w1 = -1 */
            pp=FALSE;
            for (i=1;i<e;i++)         
            { /* check for composite modulus */
                if (mr_mip->user!=NULL) (*mr_mip->user)();

                if (compare(mr_mip->w4,mr_mip->w1)==0) pp=TRUE;
                nres_modmult(mr_mip->w4,mr_mip->w4,mr_mip->w4);
                if (!pp && compare(mr_mip->w4,mr_mip->w3)==0)
                {              
                    zero(w);
                    MR_OUT
                    return FALSE;
                }
            }   
            if (compare(mr_mip->w4,mr_mip->w1)==0) break;
            if (!pp)
            {
                zero(w);
                MR_OUT
                return FALSE;
            }
        }                                                /* w= y    */
        copy(mr_mip->w15,mr_mip->w3);                    /* w3 = x  */
        nres_powmod(mr_mip->w3,mr_mip->w14,mr_mip->w15);  /* w15 = w */
        incr(mr_mip->w14,1,mr_mip->w14);
        subdiv(mr_mip->w14,2,mr_mip->w14);
        nres_powmod(mr_mip->w3,mr_mip->w14,mr_mip->w14);  /* w14 = v */
        forever
        {
            if (mr_mip->user!=NULL) (*mr_mip->user)();

            convert(1,mr_mip->w1);
            nres(mr_mip->w1,mr_mip->w1);
            if (compare(w,mr_mip->w1)==0)  break;
            copy(mr_mip->w15,mr_mip->w2);
            for (n=0;compare(mr_mip->w2,mr_mip->w1)!=0;n++)
                nres_modmult(mr_mip->w2,mr_mip->w2,mr_mip->w2);
            if (n>=e)
            {
                zero(w);
                MR_OUT
                return FALSE;
            }
            r=e-n-1;
            for (i=0;i<r;i++) nres_modmult(w,w,w);
            nres_modmult(mr_mip->w14,w,mr_mip->w14);
            nres_modmult(w,w,w);
            nres_modmult(mr_mip->w15,w,mr_mip->w15);
            e=n;
        }
        copy(mr_mip->w14,w);
        nres_modmult(w,w,mr_mip->w14);
        MR_OUT
        if (compare(mr_mip->w14,mr_mip->w3)==0) 
            return TRUE;
        zero(w);
        return FALSE;
        
    }
    return FALSE;
}

BOOL sqroot(big x,big p,big w)
{ /* w = sqrt(x) mod p */
    if (mr_mip->ERNUM) return FALSE;

    MR_IN(101)

    if (subdivisible(p,2))
    { /* p must be prime */
        zero(w);
        MR_OUT
        return FALSE;
    }
    prepare_monty(p);
    nres(x,mr_mip->w15);
    if (nres_sqroot(mr_mip->w15,w))
    {
        redc(w,w);
        MR_OUT
        return TRUE;
    }
    zero(w);
    MR_OUT
    return FALSE;
}
