// Schoof's Original Algorithm!
// Mike Scott June 1999   mike@compapp.dcu.ie
// Basic algorithm is due to Schoof
// "Elliptic Curves Over Finite Fields and the Computation of Square Roots 
// mod p", Rene Schoof, Math. Comp., Vol. 44 pp 483-494
// Expression for Mod 2 Cardinality from "Counting Points on Elliptic
// Curves over Finite Fields", Rene Schoof, Jl. de Theorie des Nombres
// de Bordeaux 7 (1995) pp 219-254
// Another useful reference is
// Elliptic Curve Public Key Cryptosystems", Menezes, 
// Kluwer Academic Publishers, Chapter 7
// Thanks are due to Richard Crandall for the tip about using prime powers
//
// **
// This program implements Schoof's original algorithm, augmented by
// the use of prime powers. By finding the Number of Points mod the product
// of many small primes and large prime powers, the final search for NP is 
// greatly speeded up.
// Final phase search uses Pollard Lambda ("kangaroo") algorithm
// This final phase effectively stretches the range of Schoof's
// algorithm by about 70-80 bits.
// This approach is only feasible due to the use of fast FFT methods
// for large degree polynomial multiplication
// **
//
// Ref "Monte Carlo Methods for Index Computation"
// by J.M. Pollard in Math. Comp. Vol. 32 1978 pp 918-924
// Ref "A New Polynomial Factorisation Algorithm and its implementation",
// Victor Shoup, Jl. Symbolic Computation, 1996 
//
// This is very much a first pass at the problem.
// A big improvement in performance will be obtained when I manage to
// replace the division polynomials with the much smaller Modular Polynomials
// Any help, advice, enlightenment would be gratefully received!
//
// This implementation does NOT use state-of-the-art ideas, and is quite slow
// by modern standards. A faster program will be made available if/when I get
// to grips with afore-mentioned ideas. Nonetheless this program is quite
// adequate for counting points on curves over GF(p), for p<= 256 bits in 
// length.
//
// It only works at all due to the use of fast FFT methods for polynomial
// arithmetic. Since Polynomial GCDs are hard to speed up, they are avoided.
// For this reason the "eigenvalue" heuristic, originally implemented, has 
// been dropped.
//
// This implementation is free. No terms, no conditions. It requires 
// version 4.22 or greater of the MIRACL library (a Shareware, Commercial 
// product, but free for non-profit making use), 
// available from ftp://ftp.compapp.dcu.ie/pub/crypto/miracl.zip 
//
// However this program may be used (unmodified) to generate curves for 
// commercial use.
//
// 32-bit build only
//
// Revision history
//
// Rev. 1 June 1999   - included prime powers
// Rev. 2 June 1999   - tweaked some inner loops
// Rev. 3 July 1999   - changed from rational to projective coordinates
//                      power windowing helps X^PP, Y^PP calculations
// Rev. 4 August 1999 - suppressed unnecessary creation of ZZn temporaries 
//                      in poly.cpp and polymod.cpp
//
// Timings for test curve Y^2=X^3-3X+49 mod 65112*2^144-1 (160 bits)
// Pentium Pro 200MHz
//
// Rev. 0 - 100 minutes
// Rev. 1 - 67 minutes
// Rev. 2 - 57 minutes
// Rev. 3 - 51 minutes
// Rev. 4 - 46 minutes
//
// Copyright Shamus Software Ltd. 1999
//

#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <string.h>
#include <elliptic.h>    // Elliptic Curve Class
#include <crt.h>         // Chinese Remainder Theorem Class

// poly.h implements polynomial arithmetic. FFT methods are used for maximum
// speed, as the polynomials get very big. For example when searching for the
// curve cardinality mod the prime 31, the polynomials are of degree (31*31-1)/2
// = 480. But all that gruesome detail is hidden away.
//
// polymod.h implements polynomial arithmetic wrt to a preset poynomial 
// modulus. This looks a bit neater. Function setmod() sets the modulus 
// to be used. Again fast FFT methods are used.

#include "poly.h"
#include "polymod.h"

Miracl precision=20;            // max. 20x32 bits per big number

PolyMod MY2,MY4;

ZZn A,B;         // Here ZZn are integers mod the prime p
                 // Montgomery representation is used internally

// Elliptic curve Point duplication formula

void elliptic_dup(PolyMod& X,PolyMod& Y,PolyMod& Z)
{ // (X,Y,Z)=2.(X,Y,Z)
    PolyMod W1,W2,W3,W4;

    W2=Z*Z;           // 1
    W3=A*(W2*W2);     // 2
    W1=X*X;           // 3
    W4=3*W1+W3;
    Z=2*Z*Y;          // 4   Z has an implied y
    W2=MY2*(Y*Y);     // 5
    W3=4*X*W2;        // 6
    W1=W4*W4;         // 7
    X=W1-2*W3;
    W2=8*(W2*W2);     // 8
    W3-=X;
    W3=W3*W4;         // 9
    Y=W3-W2;    
    X=MY2*X;          // fix up - move implied y from Z to Y
    Y=MY2*Y;
    Z=MY2*Z;
}

//
// This is addition formula for two distinct points on an elliptic curve
// Works with projective coordinates which are automatically reduced wrt a 
// polynomial modulus
// Remember that the expression for the Y coordinate of each point 
// (a function of X) is implicitly multiplied by Y.
// We know Y^2=X^3+AX+B, but we don't have an expression for Y
// So if Y^2 ever crops up - substitute for it 

void elliptic_add(PolyMod& XT,PolyMod& YT,PolyMod& ZT,PolyMod& X,PolyMod& Y)
{ // add (X,Y,1) to (XT,YT,ZT)
  // on an elliptic curve 
    PolyMod W1,W2,W4,W5,W6;

    W1=XT;
    W2=YT;          // W2 has an implied y
    W6=ZT*ZT;       // 1
    W4=X*W6;        // 2
    W6=W6*ZT;       // 3
    W5=Y*W6;        // 4 W5 has an implied y
    W1-=W4;
    W2-=W5;
    if (iszero(W1))
    {
        if (iszero(W2)) 
        { // should have doubled
            elliptic_dup(XT,YT,ZT);
            return;
        }
        else
        { // point at infinity
            ZT.clear();
            return;    
        }
    }

    W4=W1+2*W4;     // W4=2*W4+W1 
    W5=W2+2*W5;     // W5=2*W5+W2

    ZT=ZT*W1;       // 5

    W6=W1*W1;       // 6
    W1=W1*W6;       // 7
    W6=W6*W4;       // 8
    W4=MY2*(W2*W2);   // 9 Substitute for Y^2

    XT=W4-W6;

    W6=W6-2*XT;
    W2=W2*W6;       // 10
    W1=W1*W5;       // 11
    W5=W2-W1;

    YT=W5/(ZZn)2;   
}

//
//   Program to compute the order of a point on an elliptic curve
//   using Pollard's lambda method for catching kangaroos. 
//
//   As a way of counting points on an elliptic curve, this
//   has complexity O(p^(1/4))
//
//   However Schoof puts a spring in the step of the kangaroos
//   allowing them to make bigger jumps, and lowering overall complexity
//   to O(p^(1/4)/sqrt(L)) where L is the product of the Schoof primes
//
//   See "Monte Carlo Methods for Index Computation"
//   by J.M. Pollard in Math. Comp. Vol. 32 1978 pp 918-924
//

Big kangaroo(Big p,Big order,Big ordermod)
{
    ECn ZERO,X,P,G,table[50],trap;
    Big xc,mean,leaps,upper,lower,a,b,x,y,n,w,t,nrp;
    int i,j,m,sp;
    Big dm,k,dn,s,distance[50];
    BOOL bad;

    forever
    {
// find a random point on the curve
        do
        {
            x=rand(p);
        } while (!P.set(x,x));

        lower=p+1-2*sqrt(p)-3; // lower limit of search
        upper=p+1+2*sqrt(p)+3; // upper limit of search

        if (ordermod>1)
            upper+=(ordermod+order-upper%ordermod);
        w=1+(upper-lower)/ordermod;
        leaps=sqrt(w);
        mean=leaps/4;      // ideal mean for set S=1/4*w^(0.5)

        for (s=1,m=1;;m++)
        { /* find table size */
            distance[m-1]=s*ordermod;
            s*=2;
            if ((2*s/m)>mean) break;
        }

        table[0]=ordermod*P;
        for (i=1;i<m;i++)
        { // double last entry
            table[i]=table[i-1];
            table[i]+=table[i-1];
        }

        X=upper*P;
        cout << "Setting trap for Kangaroo.... \n";

        cout << "Percentage Complete    " << flush;
        for (dn=0,k=0;k<leaps;k+=1)
        { /* set traps beyond range using tame kangaroo */
            X.get(xc);
            i=xc%m;  /* random function */
            X+=table[i];
            dn+=distance[i];
            if (k%1000==0)
            {
                cout << "\b\b\b";
                cout << setw(3) << (100*k)/leaps << flush;
            }
        }

        cout << "\b\b\b100";
        cout << "\nReleasing Kangaroo....\n";
        cout << "Final Countdown    ";
        trap=X;
        X=ZERO;
        for (dm=0,j=0;;j++)
        { /* unlease wild kangaroo - boing - boing ... */
            X.get(xc);
            i=xc%m;
            X+=table[i];
            dm+=distance[i];
            if (X==trap || dm>upper+dn-lower) break;
            if (j%1000==0)
            {
                cout << "\b\b\b";
                cout << setw(3) << 100-(100*dm)/(upper+dn-lower) << flush;
            }
        }
        cout << endl;
        if (dm>upper+dn-lower)
        { /* trap stepped over */
            cout << "trap failed - must try again" << endl;
            continue;
        }
        else
        {
            nrp=upper+dn-dm;
            G=P;
            G*=nrp;
            if (G==ZERO)
            {
                if (prime(nrp))
                {
                    cout << "NP= " << nrp << endl;
                    cout << "NP is Prime!" << endl;
                    break;
                }

// final checks....
                order=nrp; i=0; 
                forever
                {
                    sp=get_mip()->PRIMES[i];
                    if (sp==0) break;
                    if (order%sp==0)
                    {
                        G=P;
                        G*=(order/sp);
                        if (G==ZERO) 
                        {
                            order/=sp;
                            continue;
                        }
                    }
                    i++;
                }   
                if (order <= 4*sqrt(p))
                { 
                    cout << "Low Order point used - trying again" << endl;
                    continue;
                }
                order=nrp;
                for (i=0;(sp=get_mip()->PRIMES[i])!=0;i++)
                    while (order%sp==0) order/=sp;
                if (order==1)
                { // all factors of nrp were considered
                    cout << "NP= " << nrp << endl;
                    break;
                }
                if (prime(order))
                { // all factors of NP except for one last one....
                    G=P;
                    G*=(nrp/order);
                    if (G==ZERO) 
                    {
                        cout << "Failed - trying again" << endl;
                        continue;
                    }
                    else 
                    {
                        cout << "NP= " << nrp << endl;
                        break;
                    }
                }

// Couldn't be bothered factoring nrp completely
// Probably not an interesting curve for Cryptographic purposes anyway.....
// But if 20 random points are all "killed" by nrp, its almost
// certain to be the true NP, and not a multiple of a small order.

                bad=FALSE;
                for (i=0;i<20;i++)
                { 
                    do
                    {
                        x=rand(p);
                    } while (!P.set(x,x));
                    G=P;
                    G*=nrp;
                    if (G!=ZERO)
                    {
                        bad=TRUE;
                        break;
                    }
                }
                if (bad)
                {
                    cout << "Failed - trying again" << endl;
                    continue;                       
                }
                cout << "NP is composite and not ideal for Cryptographic use" << endl;
                cout << "NP= " << nrp << " (probably)" << endl;
                break;
            }
            else cout << "Failed - trying again" << endl;
        }
    }
    return nrp;
}

// Code to parse formula in command line
// This code isn't mine, but its public domain
// Shamefully I forget the source

Big tt;
static char *ss;

void eval_power (Big& oldn,Big& n,char op)
{
        int i;
        if (op) n=pow(oldn,toint(n));    // power(oldn,size(n),n,n);
}

void eval_product (Big& oldn,Big& n,char op)
{
        switch (op)
        {
        case '*':
                n*=oldn; 
                break;
        case '/':
                n=oldn/n;
                break;
        case '%':
                n=oldn%n;
        }
}

void eval_sum (Big& oldn,Big& n,char op)
{
        switch (op)
        {
        case '+':
                n+=oldn;
                break;
        case '-':
                n=oldn-n;
        }
}

void eval (void)
{
        Big oldn[3];
        Big n;
        int i;
        char oldop[3];
        char op;
        char minus;
        for (i=0;i<3;i++)
        {
            oldop[i]=0;
        }
LOOP:
        while (*ss==' ')
        ss++;
        if (*ss=='-')    /* Unary minus */
        {
        ss++;
        minus=1;
        }
        else
        minus=0;
        while (*ss==' ')
        ss++;
        if (*ss=='(' || *ss=='[' || *ss=='{')    /* Number is subexpression */
        {
        ss++;
        eval ();
        n=tt;
        }
        else            /* Number is decimal value */
        {
        for (i=0;ss[i]>='0' && ss[i]<='9';i++)
                ;
        if (!i)         /* No digits found */
        {
                cout <<  "Error - invalid number" << endl;
                exit (20);
        }
        op=ss[i];
        ss[i]=0;
        n=atol(ss);
        ss+=i;
        *ss=op;
        }
        if (minus) n=-n;
        do
        op=*ss++;
        while (op==' ');
        if (op==0 || op==')' || op==']' || op=='}')
        {
        eval_power (oldn[2],n,oldop[2]);
        eval_product (oldn[1],n,oldop[1]);
        eval_sum (oldn[0],n,oldop[0]);
        tt=n;
        return;
        }
        else
        {
        if (op=='#')
        {
                eval_power (oldn[2],n,oldop[2]);
                oldn[2]=n;
                oldop[2]='#';
        }
        else
        {
                if (op=='*' || op=='/' || op=='%')
                {
                eval_power (oldn[2],n,oldop[2]);
                oldop[2]=0;
                eval_product (oldn[1],n,oldop[1]);
                oldn[1]=n;
                oldop[1]=op;
                }
                else
                {
                if (op=='+' || op=='-')
                {
                        eval_power (oldn[2],n,oldop[2]);
                        oldop[2]=0;
                        eval_product (oldn[1],n,oldop[1]);
                        oldop[1]=0;
                        eval_sum (oldn[0],n,oldop[0]);
                        oldn[0]=n;
                        oldop[0]=op;
                }
                else    /* Error - invalid operator */
                {
                        cout <<  "Error - invalid operator" << endl;
                        exit (20);
                }
                }
        }
        }
        goto LOOP;
}

int qpow(int x,int y)
{ // quick and dirty power function
    int r=x;
    for (int i=1;i<y;i++) r*=x;
    return r;
}

int main(int argc,char **argv)
{
    ofstream ofile;
    int lower,ip,pbits,lp,i,j,m,n,nl,L,*l,t[100],k,tau;
    Big a,b,p,nrp,x,y,d,s;
    PolyMod XX,XP,XPP,YP,YPP;
    Poly G,P[200],P2[200],P3[200],Y2,Y4;
    miracl *mip=&precision;
    BOOL escape,search,fout,dir,gotP,gotA,gotB;

    argv++; argc--;
    if (argc<1)
    {
        cout << "Incorrect Usage" << endl;
        cout << "Program finds the number of points (NP) on an Elliptic curve" << endl;
        cout << "which is defined over the Galois field GF(P), P a prime" << endl;
        cout << "The Elliptic Curve has the equation Y^2 = X^3 + AX + B mod P" << endl;
        cout << "schoof <prime number P> <A> <B>" << endl;
        cout << "OR" << endl;
        cout << "schoof -f <formula for P> <A> <B>" << endl;
        cout << "e.g. schoof -f 2#192-2#64-1 -3 35317045537" << endl;
        cout << "To output to a file, use flag -o <filename>" << endl;
        cout << "To search downwards for a prime, use flag -d" << endl;
        cout << "To search for NP prime, incrementing B, use flag -s" << endl;
        cout << "\nFreeware from Shamus Software, Dublin, Ireland" << endl;
        cout << "Full C++ source code and MIRACL multiprecision library available" << endl;
        cout << "http://indigo.ie/~mscott for details" << endl;
        cout << "or email mscott@indigo.ie" << endl;
        return 0;
    }

    ip=0;
    gprime(10000);                 // generate small primes < 1000
    search=fout=dir=gotP=gotA=gotB=FALSE;
    p=0; a=0; b=0;

// Interpret command line

    while (ip<argc)
    {
        if (strcmp(argv[ip],"-f")==0)
        {
            ip++;
            if (!gotP && ip<argc)
            {

                ss=argv[ip++];
                tt=0;
                eval();
                p=tt;
                gotP=TRUE;
                continue;
            }
            else
            {
                cout << "Error in command line" << endl;
                return 0;
            }
        }
        if (strcmp(argv[ip],"-o")==0)
        {
            ip++;
            if (ip<argc)
            {
                fout=TRUE;
                ofile.open(argv[ip++]);
                continue;
            }
            else
            {
                cout << "Error in command line" << endl;
                return 0;
            }
        }
        if (strcmp(argv[ip],"-d")==0)
        {
            ip++;
            dir=TRUE;
            continue;
        }
        if (strcmp(argv[ip],"-s")==0)
        {
            ip++;
            search=TRUE;
            continue;
        }
        if (!gotP) 
        {
            p=argv[ip++];
            gotP=TRUE;
            continue;
        }
        if (!gotA) 
        {
            a=argv[ip++];
            gotA=TRUE;
            continue;
        }
        if (!gotB) 
        {
            b=argv[ip++];
            gotB=TRUE;
            continue;
        }
        cout << "Error in command line" << endl;
        return 0;
    }    

    if (!gotP || !gotA || !gotB)
    {
        cout << "Error in command line" << endl;
        return 0;
    }

    if (!prime(p))
    {
        int incr=0;
        cout << "That number is not prime!" << endl;
        if (dir)
        {
            cout << "Looking for next lower prime" << endl;
            p-=1; incr++;
            while (!prime(p)) { p-=1;  incr++; }
            cout << "Prime P = P-" << incr << endl;
        }
        else
        {
            cout << "Looking for next higher prime" << endl;
            p+=1; incr++;
            while (!prime(p)) { p+=1;  incr++; }
            cout << "Prime P = P+" << incr << endl;
        }
        cout << "Prime P = " << p << endl;
    }
    pbits=bits(p);
    cout << "P mod 8 = " << p%8 << endl;
    cout << "P is " << pbits << " bits long" << endl;

// loop for "-s" search option

    forever {

    fft_reset();   // reset FFT tables

    ecurve(a,b,p,MR_AFFINE);    // initialise Elliptic Curve

    A=a;
    B=b;

// The elliptic curve as a Polynomial

    Y2=0;
    Y2.addterm(B,0);
    Y2.addterm(A,1);
    Y2.addterm((ZZn)1,3);
    Y4=Y2*Y2;

    cout << "Counting the number of points (NP) on the curve" << endl;
    cout << "y^2= " << Y2 << " mod " << p << endl;   

    if ((4*a*a*a+27*b*b)%p==0)
    {
        cout << "Not Allowed! 4A^3+27B^2 = 0" << endl;
        if (search) {b+=1; continue; }
        else break;
    }
 
    if (pbits<12)
    { // do it the simple way
        nrp=1;
        x=0;
        while (x<p)
        {
            nrp+=1+jacobi((x*x*x+a*x+b)%p,p);
            x+=1;
        }
        cout << "NP= " << nrp << endl;
        if (prime(nrp)) cout << "NP is Prime!" << endl;
        else if (search) {b+=1; continue; }
        break;
    } 
    if (pbits<56) 
    {
        nrp=kangaroo(p,(Big)0,(Big)1);
        if (!prime(nrp) && search) {b+=1; continue; }
        break;;
    }

    d=pow((Big)2,72);
    d=sqrt(p/d);
    if (d<256) d=256;

    int l[100],pp[100];   // primes and powers 

// see how many primes will be needed
// l[.] is the prime, pp[.] is the power

    for (s=1,nl=0;s<=d;nl++)
    {
        int tp=mip->PRIMES[nl];
        pp[nl]=1;     // every prime included once...
        s*=tp;
        for (i=0;i<nl;i++)
        { // if a new prime power is now in range, include its contribution
            int cp=mip->PRIMES[i];
            int p=qpow(cp,pp[i]+1);
            if (p<tp)
            { // new largest prime power
                s*=cp;
                pp[i]++;
            } 
        }
    }  
    L=mip->PRIMES[nl-1];

    cout << nl << " primes used (plus largest prime powers), largest is " << L << endl;

    for (i=0;i<nl;i++)
        l[i]=mip->PRIMES[i];

    int start_prime;   // start of primes & largest prime powers
    for (i=0;;i++)
    {
        if (pp[i]!=1)
        {
            int p=qpow(l[i],pp[i]);
            for (j=0;l[j]<p;j++) ;
            nl++;
            for (m=nl-1;m>j;m--) l[m]=l[m-1];
            l[j]=p;   // insert largest prime power in table
        }
        else 
        {
            start_prime=i;
            break;
        } 
    }

// table of primes and prime powers now looks like:-
// 2 3 5 7 9 11 13 16 17 19 ....
//     S   p        p

// where S is start_prime, and p marks the largest prime powers in the range
// CRT uses primes starting from S, but small primes are kept in anyway, 
// as they allow quick abort if searching for prime NP.

// Calculate Divisor Polynomials - Schoof 1985 p.485
// Set the first few by hand....
    
    P[1]=1; P[2]=2; P[3]=0; P[4]=0;

    P2[2]=P[2]*P[2];
    P3[2]=P2[2]*P[2];

    P[3].addterm(-(A*A),0); P[3].addterm(12*B,1);
    P[3].addterm(6*A,2)   ; P[3].addterm((ZZn)3,4);

    P2[3]=P[3]*P[3];
    P3[3]=P2[3]*P[3];

    P[4].addterm((ZZn)(-4)*(8*B*B+A*A*A),0);
    P[4].addterm((ZZn)(-16)*(A*B),1);
    P[4].addterm((ZZn)(-20)*(A*A),2);
    P[4].addterm((ZZn)80*B,3);
    P[4].addterm((ZZn)20*A,4);
    P[4].addterm((ZZn)4,6);

    P2[4]=P[4]*P[4];
    P3[4]=P2[4]*P[4];

    lower=5;     // next one to be calculated

  // Finding the order modulo 2
  // If GCD(X^P-X,X^3+AX+B) == 1 , trace=1 mod 2, else trace=0 mod 2

    XX=0;
    XX.addterm((ZZn)1,1);

    setmod(Y2);

    XP=pow(XX,p);
    G=gcd(XP-XX);         
    t[0]=0;
    if (isone(G)) t[0]=1;
    cout << "NP mod 2 =   " << (p+1-t[0])%2;
    if ((p+1-t[0])%2==0)
    {                                  
        cout << " ***" << endl;
        if (search) {b+=1; continue; }
    }
    else cout << endl;

    Poly zero;
    PolyMod one,XT,YT,ZT;
    one=1;                  // polynomial = 1
    zero=0;                 // polynomial = 0

    Crt CRT(nl-start_prime,&l[start_prime]);  // initialise for application of the 
                                              // chinese remainder thereom

 // now look for trace%prime for prime=3,5,7,11 etc
 // actual trace is found by combining these via CRT

    escape=FALSE;
    for (i=1;i<nl;i++)      
    {
        lp=l[i];       // next prime
        k=p%lp;

// generation of Divisor polynomials as needed
// See Schoof p. 485

        for (j=lower;j<=lp+1;j++)
        { // different for even and odd 
            if (j%2==1)
            { 
                n=(j-1)/2;
                if (n%2==0)
                    P[j]=P[n+2]*P3[n]*Y4-P3[n+1]*P[n-1];
                else
                    P[j]=P[n+2]*P3[n]-Y4*P3[n+1]*P[n-1];

                if (j <= 1+(L+1)/2)
                { // precalculate squares and cubes for later
                    P2[j]=P[j]*P[j];
                    P3[j]=P2[j]*P[j];
                }
            }
            else
            {
                n=j/2;
                P[j]=P[n]*(P[n+2]*P2[n-1]-P[n-2]*P2[n+1])/(ZZn)2;
         
                if (j <= 1+(L+1)/2)
                { // precalculate for later
                    P2[j]=P[j]*P[j];
                    P3[j]=P2[j]*P[j];
                }
            }
        }

        lower=lp+2;

        setmod(P[lp]);
        MY2=Y2;
        MY4=Y4;

// These next are time-consuming calculations of X^P, X^(P*P), Y^P and Y^(P*P)

        cout << "X^P " << flush;
        XP=pow(XX,p);
        cout << "\b\b\b\bY^P " << flush;
        YP=pow(MY2,(p-1)/2);
        YP=zero-YP;
        cout << "\b\b\b\bX^PP" << flush;
        XPP=pow(XP,p);
        cout << "\b\b\b\bY^PP" << flush;
        YPP=pow(YP,(p+1));
        cout << "\b\b\b\b";

        PolyMod Pk,P2k,PkP1,PkM1,PkP2;
        Pk=P[k]; PkP1=P[k+1]; PkM1=P[k-1]; PkP2=P[k+2];

        P2k=(Pk*Pk);
//
// This is Schoof's algorithm, stripped to its bare essentials
//
// Now looking for the value of tau which satisfies 
// (X^(P*P),Y^(P*P)) + k.(X,Y) =  tau.(X^P,Y^P)
// 
// Calculate LHS of this expression first. If its zero, then tau=0
// Note that (X,Y) are rational polynomial expressions for points on
// an elliptic curve, so "+" means elliptic curve point addition
// 
// k.(X,Y) can be found directly from Divisor polynomials
// Schoof Prop (2.2)
//
// Points are converted to projective (X,Y,Z) form
// This is faster (x2). Observe that (X/Z^2,Y/Z^3,1) is the same
// point in projective co-ordinates as (X,Y,Z)
//

        if (k%2==0)
        {
            XT=XX*MY2*P2k-PkM1*PkP1;
            YT=(PkP2*PkM1*PkM1-P[k-2]*PkP1*PkP1)/4;
            XT=MY2*XT;      // fix up, so that Y has implicit y multiplier
            YT=MY2*YT;      // rather than Z
            ZT=MY2*Pk;
        }
        else
        {
            XT=(XX*P2k-MY2*PkM1*PkP1);
            if (k==1) YT=(PkP2*PkM1*PkM1+PkP1*PkP1)/4;
            else      YT=(PkP2*PkM1*PkM1-P[k-2]*PkP1*PkP1)/4;
            ZT=Pk;
        }

        elliptic_add(XT,YT,ZT,XPP,YPP);

// Test for Schoof's case 1

        cout << "NP mod " << lp << " = " << flush;

        if (iszero(ZT))
        { // Is it zero point? (XPP,YPP) = - K(X,Y)
            t[i]=0;
            cout << setw(3) << (p+1)%lp;
            if ((p+1)%lp==0)
            {      
                cout << " ***" << endl;
                if (search) {escape=TRUE; break;}
            }
            else cout << endl;
            continue;         
        }

// Schoof's Case 2
// Now keep adding (X^P,-Y^P), until result is point at infinity
// This is very simple!
//
        for (tau=1;tau<lp;tau++)
        { 
            cout << setw(3) << (p+1-tau)%lp << flush;     
            elliptic_add(XT,YT,ZT,XP,YP);
            if (iszero(ZT))
            { // Is it zero point?
                t[i]=tau;
                if ((p+1-tau)%lp==0)
                {
                    cout << " ***" << endl;
                    if (search) escape=TRUE;
                }
                else cout << endl;
                break;
            }
            cout << "\b\b\b";            
        }
        if (escape) break;
    }

    for (i=0;i<=L+1;i++) 
    {
        P[i].clear(); // reclaim space
        P2[i].clear();
        P3[i].clear();
    }

    if (escape) {b+=1; continue;}

    Big order,ordermod;
    ordermod=1; for (i=0;i<nl-start_prime;i++) ordermod*=l[start_prime+i];
    order=(p+1-CRT.eval(&t[start_prime]))%ordermod;    // get order mod product of primes

    nrp=kangaroo(p,order,ordermod);

    if (!prime(nrp) && search) {b+=1; continue; }
    else break;
    }

    if (fout) 
    {
        ECn P;
        ofile << a << endl;
        ofile << b << endl;
    // generate a random point on the curve 
        do {
            x=rand(p);
        } while (!P.set(x,x));
        P.get(x,y);
        ofile << x << endl;
        ofile << y << endl;
        ofile << p << endl;
        ofile << nrp << endl;
    }
    return 0;
}

