// Schoof-Elkies-Atkin Algorithm!
// Mike Scott August 1999   mike@compapp.dcu.ie
// Counts points on GF(p) Elliptic Curve, y^2=x^3+Ax+B a prerequisite 
// for implemention of  Elliptic Curve Cryptography
// This program is intended as an aid to Cryptographers in generating 
// suitable curves (ideally with a prime number of points) with respect 
// to prime moduli from 160 to 512 bits in length. This can be done with
// relatively modest computing facilities - the average home computer will
// suffice.
// An "ideal" curve is defined as one with with prime number of points.
//
// First the credits
//
// Basic algorithm is due to Schoof
// 1. "Elliptic Curves Over Finite Fields and the Computation of Square Roots 
//     mod p", Rene Schoof, Math. Comp., Vol. 44 pp 483-494
// 
// Elkies-Atkin ideas are described in
//
// 2. "Counting points on Elliptic Curves over Finite Fields", Rene Schoof, 
//    Jl. de Theorie des Nombres de Bordeaux 7 (1995) pp 219-254
//
// The particular variation implemented here is due to Mueller. See his thesis
//
// 3. "Ein Algorithmus zur Bestimmung der Punktanzahl elliptischer Kurven
//    uber endlichen Korpern der Charakteristik grosser drei",  
//    available from Volker Mueller's home page 
//    www.informatik.th-darmstadt.de/TI/Mitarbeiter/vmueller.html
//
// Other useful English-language publications are available from this site.
// Strongly recommended is the recent book 
//
// 4. "Elliptic Curves in Cryptography"
//     by Blake, Seroussi and Smart, London Mathematical Society Lecture Note 
//     Series 265, Cambridge University Press. ISBN 0 521 65374 6 
//
// Another useful reference is
// 5.  Elliptic Curve Public Key Cryptosystems", Menezes, 
//     Kluwer Academic Publishers, Chapter 7
//
// The Kangaroo continuation is due to Pollard:-
//
// 6. "Monte Carlo Methods for Index Computation"
//    by J.M. Pollard in Math. Comp. Vol. 32 1978 pp 918-924
//
// Fast FFT methods are largely as described by Shoup
//
// 7. "A New Polynomial Factorisation Algorithm and its implementation",
//     Victor Shoup, Jl. Symbolic Computation, 1996 
//
// Thanks are due to Richard Crandall for his encouragment, and the idea of 
// using projective co-ordiantes.
//
// NOTE: Only for use on curves over large prime modulus P. 
// For smaller P use schoof.exe utility available from the same source.
//
// This first version does not process so-called Atkin primes
// Schoof's original algorithm is used for primes 3, 5 and 7 (this facilitates
// "early abort" using the -s option).
// After that only Elkies primes are used. The final continuation uses Pollard's
// Lambda ("kangaroo") algorithm. It is therefore not as fast as it could
// be, particularly for smaller values of the prime modulus P. However when 
// the asymptotics kick-in, it starts to get competitive (when you need it 
// to).
//
// It is envisaged that from time-to-time this program will be modified
// and hopefully improved - that is speeded up.
// In particular it is planned to exploit Atkin Primes to allow two herds of 
// kangaroos complete the job more efficiently
//
// Asyptotically the algorithm should take time approx. O(log(p)^4)
// Timings:- 3 random curves 180MHz Pentium Pro
//
//               C1      C2      C3    Asymptotic multiplier wrt 160 bits
// 160-bit      515     550     345                 1
// 192-bit      880    1655     885                 2
// 256 bit     3200                                 6.5
// 384 bit    19800                                 33
//
// As can be seen the asymptotic behaviour of the program would appear to 
// be correct.
//
// ****************
// Instructions for use
//
// First run the utility "mueller" to build up a collection of Modular 
// Polynomials. This needs to be done once only - ever, but you can from time 
// augment your collection of Polynomials by running it again. Its quite time
// consuming, but in less than an hour you should have enough to get started. The
// more you have, the bigger the prime modulus that you can use.
//
// Then run the utility "process" to process the raw polynomial file with 
// respect to your chosen prime modulus. This need to be done just once for
// every prime modulus of interest to you. This takes only a few minutes at 
// most.
//
// An alternative is to use instead the "modpol" application, which is a
// composite of "mueller"  and "process". It directly generates the Modular 
// Polynomials reduced wrt a pre-specified prime modulus, suitable for 
// immediate use by "sea". If working with limited computing resources such 
// that sufficient generic Modular Polynomials cannot be generated by 
// "mueller", this may be your only alternative.
//
// Finally run this program "sea" specifying the A and B parameters of the 
// particular curve. This program can also search through many curves for 
// a curve ideal for cryptographic use (with a prime number of points).
//
// For example try:-
//
// mueller 0 120 -o mueller.raw
// process -f 65112*2#144-1 -i mueller.raw -o test160.pol
// sea -3 49 -i test160.pol
//
// or

// modpol -f 65112*2#144-1 0 120 -o test160.pol
// sea -3 49 -i test160.pol
//
//
// When using the "sea" program, the -s option is particularly useful
// and allows automatic search for an "ideal" curve. If a curve order is
// exactly divisible by a small prime, that curve is immediately abandoned, 
// and the program moves on to the next, incrementing the B parameter of 
// the curve. 
//
// ****************
//
// This implementation is free. No terms, no conditions. It requires 
// version 4.23 or greater of the MIRACL library (a Shareware, Commercial 
// product, but free for non-profit making use), 
// available from ftp://ftp.compapp.dcu.ie/pub/crypto/miracl.zip 
//
// However this program may be freely used (unmodified!) to generate curves 
// for commercial use. It may be recompiled for this purpose on any hardware.
//
// 32-bit build only
//
// Copyright Shamus Software Ltd. 1999
//

#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>
#include <string.h>
#include <elliptic.h>    // Elliptic Curve Class
#include <crt.h>         // Chinese Remainder Theorem Class

//
// poly.h implements polynomial arithmetic. FFT methods are used for maximum
// speed, as the polynomials can get very big. 
// But all that gruesome detail is hidden away.
//
// polymod.h implements polynomial arithmetic wrt to a preset poynomial 
// modulus. This looks a bit neater. Function setmod() sets the modulus 
// to be used. Again fast FFT methods are used.
//
// polyxy.h implements a bivariate polynomial class
//

#include "poly.h"
#include "polymod.h"
#include "polyxy.h"

Miracl precision=40;            // max. 40x32 bits per big number

PolyMod MY2,MY4;

ZZn A,B;         // Here ZZn are integers mod the prime p
                 // Montgomery representation is used internally

// Elliptic curve Point duplication formula

void elliptic_dup(PolyMod& X,PolyMod& Y,PolyMod& Z)
{ // (X,Y,Z)=2.(X,Y,Z)
    PolyMod W1,W2,W3,W4;

    W2=Z*Z;           // 1
    W3=A*(W2*W2);     // 2
    W1=X*X;           // 3
    W4=3*W1+W3;
    Z=2*Z*Y;          // 4   Z has an implied y
    W2=MY2*(Y*Y);     // 5
    W3=4*X*W2;        // 6
    W1=W4*W4;         // 7
    X=W1-2*W3;
    W2=8*(W2*W2);     // 8
    W3-=X;
    W3=W3*W4;         // 9
    Y=W3-W2;    
    X=MY2*X;          // fix up - move implied y from Z to Y
    Y=MY2*Y;
    Z=MY2*Z;
}

//
// This is addition formula for two distinct points on an elliptic curve
// Works with projective coordinates which are automatically reduced wrt a 
// polynomial modulus
// Remember that the expression for the Y coordinate of each point 
// (a function of X) is implicitly multiplied by Y.
// We know Y^2=X^3+AX+B, but we don't have an expression for Y
// So if Y^2 ever crops up - substitute for it 

void elliptic_add(PolyMod& XT,PolyMod& YT,PolyMod& ZT,PolyMod& X,PolyMod& Y)
{ // add (X,Y,1) to (XT,YT,ZT)
  // on an elliptic curve 
    PolyMod W1,W2,W4,W5,W6;

    W1=XT;
    W2=YT;          // W2 has an implied y
    W6=ZT*ZT;       // 1
    W4=X*W6;        // 2
    W6=W6*ZT;       // 3
    W5=Y*W6;        // 4 W5 has an implied y
    W1-=W4;
    W2-=W5;
    if (iszero(W1))
    {
        if (iszero(W2)) 
        { // should have doubled
            elliptic_dup(XT,YT,ZT);
            return;
        }
        else
        { // point at infinity
            ZT.clear();
            return;    
        }
    }

    W4=W1+2*W4;     // W4=2*W4+W1 
    W5=W2+2*W5;     // W5=2*W5+W2

    ZT=ZT*W1;       // 5

    W6=W1*W1;       // 6
    W1=W1*W6;       // 7
    W6=W6*W4;       // 8
    W4=MY2*(W2*W2);   // 9 Substitute for Y^2

    XT=W4-W6;

    W6=W6-2*XT;
    W2=W2*W6;       // 10
    W1=W1*W5;       // 11
    W5=W2-W1;

    YT=W5/(ZZn)2;   
}

//
//   Program to compute the order of a point on an elliptic curve
//   using Pollard's lambda method for catching kangaroos. 
//
//   As a way of counting points on an elliptic curve, this
//   has complexity O(p^(1/4))
//
//   However Schoof puts a spring in the step of the kangaroos
//   allowing them to make bigger jumps, and lowering overall complexity
//   to O(p^(1/4)/sqrt(L)) where L is the product of the Schoof primes
//

Big kangaroo(Big p,Big order,Big ordermod)
{
    ECn ZERO,X,P,G,table[50],trap;
    Big xc,mean,leaps,upper,lower,a,b,x,y,n,w,t,nrp;
    int i,j,m,sp;
    Big dm,k,dn,s,distance[50];
    BOOL bad;

    forever
    {
// find a random point on the curve
        do
        {
            x=rand(p);
        } while (!P.set(x,x));

        lower=p+1-2*sqrt(p)-3; // lower limit of search
        upper=p+1+2*sqrt(p)+3; // upper limit of search

        if (ordermod>1)
            upper+=(ordermod+order-upper%ordermod);
        w=1+(upper-lower)/ordermod;
        leaps=sqrt(w);
        mean=leaps/4;      // ideal mean for set S=1/4*w^(0.5)

        for (s=1,m=1;;m++)
        { /* find table size */
            distance[m-1]=s*ordermod;
            s*=2;
            if ((2*s/m)>mean) break;
        }

        table[0]=ordermod*P;
        for (i=1;i<m;i++)
        { // double last entry
            table[i]=table[i-1];
            table[i]+=table[i-1];
        }

        X=upper*P;
        cout << "Setting trap for Kangaroo.... \n";

        cout << "Percentage Complete    " << flush;
        for (dn=0,k=0;k<leaps;k+=1)
        { /* set traps beyond range using tame kangaroo */
            X.get(xc);
            i=xc%m;  /* random function */
            X+=table[i];
            dn+=distance[i];
            if (k%1000==0)
            {
                cout << "\b\b\b";
                cout << setw(3) << (100*k)/leaps << flush;
            }
        }

        cout << "\b\b\b100";
        cout << "\nReleasing Kangaroo....\n";
        cout << "Final Countdown    ";
        trap=X;
        X=ZERO;
        for (dm=0,j=0;;j++)
        { /* unlease wild kangaroo - boing - boing ... */
            X.get(xc);
            i=xc%m;
            X+=table[i];
            dm+=distance[i];
            if (X==trap || dm>upper+dn-lower) break;
            if (j%1000==0)
            {
                cout << "\b\b\b";
                cout << setw(3) << 100-(100*dm)/(upper+dn-lower) << flush;
            }
        }
        cout << endl;
        if (dm>upper+dn-lower)
        { /* trap stepped over */
            cout << "trap failed - must try again" << endl;
            continue;
        }
        else
        {
            nrp=upper+dn-dm;
            G=P;
            G*=nrp;
            if (G==ZERO)
            {
                if (prime(nrp))
                {
                    cout << "NP= " << nrp << endl;
                    cout << "NP is Prime!" << endl;
                    break;
                }

// final checks....
                order=nrp; i=0; 
                forever
                {
                    sp=get_mip()->PRIMES[i];
                    if (sp==0) break;
                    if (order%sp==0)
                    {
                        G=P;
                        G*=(order/sp);
                        if (G==ZERO) 
                        {
                            order/=sp;
                            continue;
                        }
                    }
                    i++;
                }   
                if (order <= 4*sqrt(p))
                { 
                    cout << "Low Order point used - trying again" << endl;
                    continue;
                }
                order=nrp;
                for (i=0;(sp=get_mip()->PRIMES[i])!=0;i++)
                    while (order%sp==0) order/=sp;
                if (order==1)
                { // all factors of nrp were considered
                    cout << "NP= " << nrp << endl;
                    break;
                }
                if (prime(order))
                { // all factors of NP except for one last one....
                    G=P;
                    G*=(nrp/order);
                    if (G==ZERO) 
                    {
                        cout << "Failed - trying again" << endl;
                        continue;
                    }
                    else 
                    {
                        cout << "NP= " << nrp << endl;
                        break;
                    }
                }

// Couldn't be bothered factoring nrp completely
// Probably not an interesting curve for Cryptographic purposes anyway.....
// But if 20 random points are all "killed" by nrp, its almost
// certain to be the true NP, and not a multiple of a small order.

                bad=FALSE;
                for (i=0;i<20;i++)
                { 
                    do
                    {
                        x=rand(p);
                    } while (!P.set(x,x));
                    G=P;
                    G*=nrp;
                    if (G!=ZERO)
                    {
                        bad=TRUE;
                        break;
                    }
                }
                if (bad)
                {
                    cout << "Failed - trying again" << endl;
                    continue;                       
                }
                cout << "NP is composite and not ideal for Cryptographic use" << endl;
                cout << "NP= " << nrp << " (probably)" << endl;
                break;
            }
            else cout << "Failed - trying again" << endl;
        }
    }
    return nrp;
}

//
// Coefficients Ck from Mueller's lemma 6.2
// 3. page 90
// 4. page 127-128
//

void get_ck(int terms,ZZn a,ZZn b,ZZn *c)
{
    int k,h;
    if (terms==0) return;
    c[1]=-a/5;
    if (terms==1) return;
    c[2]=-b/7;
    for (k=3;k<=terms;k++)
    {
        c[k]=0;
        for (h=1;h<=k-2;h++) c[k]+=c[h]*c[k-1-h];
        c[k]*=((ZZn)3/(ZZn)((k-2)*(2*k+3)));
    }
}

int main(int argc,char **argv)
{
    ofstream ofile;
    ifstream mueller;
    int max,lower,ip,pbits,lp,i,jj,is,m,n,nx,ny,nl,L;
    int schp,good[100],l[200],t[100],k,tau,lambda;
    Big a,b,c,p,nrp,x,y,d,accum;
    PolyMod XX,YY,XP,XPP,YP,YPP;
    PolyXY MP,Gl[200];
    Poly F,G,P[100],P2[100],P3[100],Y2,Y4;
    miracl *mip=&precision;
    BOOL escape,search,fout,gotI,gotA,gotB;
    ZZn j,g,qb,qc,delta,el;

    argv++; argc--;
    if (argc<1)
    {
        cout << "Incorrect Usage" << endl;
        cout << "Program finds the number of points (NP) on an Elliptic curve" << endl;
        cout << "which is defined over the Galois field GF(P), P a prime" << endl;
        cout << "The Elliptic Curve has the equation Y^2 = X^3 + AX + B mod P" << endl;
        cout << "sea <A> <B> <-i file>" << endl;
        cout << "where the input file contains pre-processed modular polynomials" << endl;
        cout << "for a particular prime modulus P" << endl;
        cout << "e.g. sea -3 35317045537 -i mueller.pol" << endl;
        cout << "To output to a file, use flag -o <filename>" << endl;
        cout << "To search for NP prime, incrementing B, use flag -s" << endl;
        cout << "\nFreeware from Shamus Software, Dublin, Ireland" << endl;
        cout << "Full C++ source code and MIRACL multiprecision library available" << endl;
        cout << "http://indigo.ie/~mscott for details" << endl;
        cout << "or email mscott@indigo.ie" << endl;
        return 0;
    }

    ip=0;
    gprime(10000);                 // generate small primes < 1000
    search=fout=gotI=gotA=gotB=FALSE;
    a=0; b=0;

// Interpret command line

    while (ip<argc)
    {
        if (!fout && strcmp(argv[ip],"-o")==0)
        {
            ip++;
            if (ip<argc)
            {
                fout=TRUE;
                ofile.open(argv[ip++]);
                continue;
            }
            else
            {
                cout << "Error in command line" << endl;
                return 0;
            }
        }
        if (!gotI && strcmp(argv[ip],"-i")==0)
        {
            ip++;
            if (ip<argc)
            {
                gotI=TRUE;
                mueller.open(argv[ip],ios::nocreate);
                if (!mueller)
                {
                    cout << "input file " << argv[ip] << " could not be opened" << endl;
                    return 0;
                }
                ip++;
                continue;
            }
            else
            {
                cout << "Error in command line" << endl;
                return 0;
            }
        }

        if (strcmp(argv[ip],"-s")==0)
        {
            ip++;
            search=TRUE;
            continue;
        }
        if (!gotA) 
        {
            a=argv[ip++];
            gotA=TRUE;
            continue;
        }
        if (!gotB) 
        {
            b=argv[ip++];
            gotB=TRUE;
            continue;
        }
        cout << "Error in command line" << endl;
        return 0;
    }    

    if (!gotI || !gotA || !gotB)
    {
        cout << "Error in command line" << endl;
        return 0;
    }

// get prime modulus from .pol file

    mueller >> p;
    pbits=bits(p);
    cout << "P= " << p << endl;
    cout << "P mod 8 = " << p%8 << endl;
    cout << "P is " << pbits << " bits long" << endl;
    cout << "Reading in pre-processed Modular Polynomials... " << endl;
//
// read in all pre-processed bivariate modular polynomials
//
    modulo(p);   // set modulus
    l[0]=2;
    max=0;
    for (i=1;;i++)
    {
        mueller >> lp;
        if (mueller.eof()) break;
        max=i;
        l[i]=lp;
        L=lp;
        Gl[i].clear();
        forever 
        {
            mueller >> c >> nx >> ny;
            Gl[i].addterm((ZZn)c,nx,ny);
            if (nx==0 && ny==0) break;
        }
    }


// loop for "-s" search option

    forever {

    fft_reset();   // reset FFT tables

    ecurve(a,b,p,MR_AFFINE);    // initialise Elliptic Curve

    A=a;
    B=b;

// The elliptic curve as a Polynomial

    Y2=0;
    Y2.addterm(B,0);
    Y2.addterm(A,1);
    Y2.addterm((ZZn)1,3);
    Y4=Y2*Y2;

    cout << "Counting the number of points (NP) on the curve" << endl;
    cout << "y^2= " << Y2 << " mod " << p << endl;   

    delta=-16*(4*A*A*A+27*B*B);

    if (delta==0)
    {
        cout << "Not Allowed! 4A^3+27B^2 = 0" << endl;
        if (search) {b+=1; continue; }
        else break;
    }
    
 // Calculate j-invariant

    j=(-1728*64*A*A*A)/delta;

    if (j==0 || j==1728)
    {
        cout << "Not Allowed! j-invariant = 0 or 1728" << endl;
        if (search) {b+=1; continue; }
        else break;
    }

// Calculate Divisor Polynomials for small primes - Schoof 1985 p.485
// Set the first few by hand....
    
    schp=7;   // use Schoof's method for primes 2 < p < schp

    P[1]=1; P[2]=2; P[3]=0; P[4]=0;

    P2[2]=P[2]*P[2];
    P3[2]=P2[2]*P[2];

    P[3].addterm(-(A*A),0); P[3].addterm(12*B,1);
    P[3].addterm(6*A,2)   ; P[3].addterm((ZZn)3,4);

    P2[3]=P[3]*P[3];
    P3[3]=P2[3]*P[3];

    P[4].addterm((ZZn)(-4)*(8*B*B+A*A*A),0);
    P[4].addterm((ZZn)(-16)*(A*B),1);
    P[4].addterm((ZZn)(-20)*(A*A),2);
    P[4].addterm((ZZn)80*B,3);
    P[4].addterm((ZZn)20*A,4);
    P[4].addterm((ZZn)4,6);

    P2[4]=P[4]*P[4];
    P3[4]=P2[4]*P[4];

// generation of Divisor polynomials 
// See Schoof p. 485

    for (jj=5;jj<=schp+1;jj++)
    { // different for even and odd 
        if (jj%2==1)
        { 
            n=(jj-1)/2;
            if (n%2==0)
                P[jj]=P[n+2]*P3[n]*Y4-P3[n+1]*P[n-1];
            else
                P[jj]=P[n+2]*P3[n]-Y4*P3[n+1]*P[n-1];

            if (jj <= 1+(schp+1)/2)
            { // precalculate squares and cubes for later
                P2[jj]=P[jj]*P[jj];
                P3[jj]=P2[jj]*P[jj];
            }
        }
        else
        {
            n=jj/2;
            P[jj]=P[n]*(P[n+2]*P2[n-1]-P[n-2]*P2[n+1])/(ZZn)2;
      
            if (jj <= 1+(schp+1)/2)
            { // precalculate for later
                P2[jj]=P[jj]*P[jj];
                P3[jj]=P2[jj]*P[jj];
            }
        }
    }

// Finding the order modulo 2
// If GCD(X^P-X,X^3+AX+B) == 1 , trace=1 mod 2, else trace=0 mod 2

    XX=0;
    XX.addterm((ZZn)1,1);
    YY=0;
    YY.addterm((ZZn)-1,0);      // Point (X,-Y)
    setmod(Y2);

    XP=pow(XX,p);
    G=gcd(XP-XX);         
    t[0]=0;
    if (isone(G)) t[0]=1;
    cout << "NP mod 2 =   " << (p+1-t[0])%2;
    if ((p+1-t[0])%2==0)
    {                                  
        cout << " ***" << endl;
        if (search) {b+=1; continue; }
    }
    else cout << endl;
    nl=1; good[0]=2;
    accum=2;           // accumulated product of good primes

    Poly zero;
    PolyMod one,XT,YT,ZT;
    one=1;         // polynomial = 1
    zero=0;        // polynomial = 0

    PolyXY dGx,dGy,dGxx,dGxy,dGyy,dGyx;
    ZZn E0b,E0bd,E2bs,E4b,E6b,Dg,Dj,Dgd,Djd,Dgs,Djs,jl,jld,p1,jd;
    ZZn E4bl,E6bl,deltal,atilde,btilde,gd,f,fd,s,Eg,Ej;
    int r,v,ld,ld1,k;
    ZZn M,cf[500],cft[500],ad,K,RF;
    Poly Fl,T,WP[500],H,X,Y,R;

    E4b=-(A/3);
    E6b=-(B/2);
    delta=(E4b*E4b*E4b-E6b*E6b)/1728;
//
// find out how many bits we are going to need
// before Kangaroos can take over
//
    d=pow((Big)2,56);
    d=sqrt(p/d);

    escape=FALSE;

//
// Schoof's original method for small primes
//

    for (i=1;;i++)
    {
        lp=mip->PRIMES[i];
        if (lp>schp) break;
        k=p%lp;

        good[nl]=lp;
        setmod(P[lp]);
        MY2=Y2;
        MY4=Y4;

// These next are time-consuming calculations of X^P, X^(P*P), Y^P and Y^(P*P)

        XP=pow(XX,p);
        YP=pow(MY2,(p-1)/2);
        YP=zero-YP;
        XPP=pow(XP,p);
        YPP=pow(YP,(p+1));

        PolyMod Pk,P2k,PkP1,PkM1,PkP2;
        Pk=P[k]; PkP1=P[k+1]; PkM1=P[k-1]; PkP2=P[k+2];

        P2k=(Pk*Pk);
//
// This is Schoof's algorithm, stripped to its bare essentials
//
// Now looking for the value of tau which satisfies 
// (X^(P*P),Y^(P*P)) + k.(X,Y) =  tau.(X^P,Y^P)
// 
// Calculate LHS of this expression first. If its zero, then tau=0
// Note that (X,Y) are rational polynomial expressions for points on
// an elliptic curve, so "+" means elliptic curve point addition
// 
// k.(X,Y) can be found directly from Divisor polynomials
// Schoof Prop (2.2)
//
// Points are converted to projective (X,Y,Z) form
// This is faster (x2). Observe that (X/Z^2,Y/Z^3,1) is the same
// point in projective co-ordinates as (X,Y,Z)
//

        if (k%2==0)
        {
            XT=XX*MY2*P2k-PkM1*PkP1;
            YT=(PkP2*PkM1*PkM1-P[k-2]*PkP1*PkP1)/4;
            XT=MY2*XT;      // fix up, so that Y has implicit y multiplier
            YT=MY2*YT;      // rather than Z
            ZT=MY2*Pk;
        }
        else
        {
            XT=(XX*P2k-MY2*PkM1*PkP1);
            if (k==1) YT=(PkP2*PkM1*PkM1+PkP1*PkP1)/4;
            else      YT=(PkP2*PkM1*PkM1-P[k-2]*PkP1*PkP1)/4;
            ZT=Pk;
        }

        elliptic_add(XT,YT,ZT,XPP,YPP);

// Test for Schoof's case 1

        cout << "NP mod " << lp << " = " << flush;

        if (iszero(ZT))
        { // Is it zero point? (XPP,YPP) = - K(X,Y)
            t[nl]=0;
            cout << setw(3) << (p+1)%lp;
            if ((p+1)%lp==0)
            {      
                cout << " ***" << endl;
                if (search) {escape=TRUE; break;}
            }
            else cout << endl;
            continue;         
        }

// Schoof's Case 2
// Now keep adding (X^P,-Y^P), until result is point at infinity
// This is very simple!
//
        for (tau=1;tau<lp;tau++)
        { 
            cout << setw(3) << (p+1-tau)%lp << flush;     
            elliptic_add(XT,YT,ZT,XP,YP);
            if (iszero(ZT))
            { // Is it zero point?
                t[nl]=tau;
                if ((p+1-tau)%lp==0)
                {
                    cout << " ***" << endl;
                    if (search) escape=TRUE;
                }
                else cout << endl;
                break;
            }
            cout << "\b\b\b";            
        }
        nl++;
        accum*=lp;
        if (escape) break;
    }

    if (!escape) for (i=1;accum<=d;i++)      
    {
        if (i>max)
        {
            cout << "Ran out of Modular Polynomials!" << endl;
            break;
        }
        lp=l[i];

        if (lp<=schp) continue;

        for (is=1;;is++)
            if (is*(lp-1)%12==0) break;

        el=lp;
        s=is;

// Get next Modular Polynomial

        MP=Gl[i];

// Evaluate bivariate polynomial at Y=j-invariant
// and use as polynomial modulus
        F=MP.F(j);
        setmod(F);
        XP=pow(XX,p);
//
// Determine "Splitting type" 
//
        G=gcd(XP-XX);
        if (isone(G)) continue;      // Atkin Prime

        if (degree(G)!=2) continue;  // pathalogical case  
//
// Good Elkies prime - so use it!
//
// First solve quadratic for a root
//
        qb=G.coeff(1);
        qc=G.coeff(0); 
        g=sqrt(qb*qb-4*qc);
        g=(-qb-g)/2;

//
// Mueller's procedure for finding the atilde, btilde and p1
// parameters of the isogenous curve
// 3. page 111
// 4. page 131-133
// First we need partial differentials of bivariate Modular Polynomial
//

        dGx=diff_dx(MP);
        dGy=diff_dy(MP);
        dGxx=diff_dx(dGx);
        dGxy=diff_dx(dGy);
        dGyx=diff_dy(dGx);
        dGyy=diff_dy(dGy);

        Eg=dGx.F(g,j);   // Evaluated at (g,j)
        Ej=dGy.F(g,j);

        Dg=g*Eg;    
        Dj=j*Ej;

        deltal=delta*pow(g,12/is)/pow(el,12);

        if (Dj==0)
        {
            E4bl=E4b/(el*el);
            atilde=-3*pow(el,4)*E4bl;
            jl=pow(E4bl,3)/deltal;
            btilde=2*pow(el,6)*sqrt((jl-1728)*deltal);
            p1=0;
        }
        else
        {
            E2bs=(-12*E6b*Dj)/(s*E4b*Dg);

            gd=-(s/12)*E2bs*g;
            jd=-E4b*E4b*E6b/delta;
            E0b=E6b/(E4b*E2bs); 

            Dgd=gd*Eg+g*(gd*dGxx.F(g,j)+jd*dGxy.F(g,j));
            Djd=jd*Ej+j*(jd*dGyy.F(g,j)+gd*dGyx.F(g,j));  
   
            E0bd=((-s*Dgd)/12-E0b*Djd)/Dj;

            E4bl=(E4b-E2bs*(12*E0bd/E0b+6*E4b*E4b/E6b-4*E6b/E4b)+E2bs*E2bs)/(el*el);

            jl=pow(E4bl,3)/deltal;
            f=pow(el,is)/g; fd=s*E2bs*f/12;

            Dgs=dGx.F(f,jl);
            Djs=dGy.F(f,jl);

            jld=-fd*Dgs/(el*Djs);
            E6bl=-E4bl*jld/jl;

            atilde=-3*pow(el,4)*E4bl;
            btilde=-2*pow(el,6)*E6bl;
            p1=-el*E2bs/2;            
        }

//
// Find factor of Division Polynomial from atilde, btilde and p1 
// Here we follow 3. p 116
// Polynomials have been modified s.t x=z^2
//
// Note that all Polynomials can be reduced mod x^(d+1),
// where d=(lp-1)/2, using modxn() function
//

        ld=(lp-1)/2;
        ld1=(lp-3)/2;

        get_ck(ld1,A,B,cf);

        WP[1]=1;
        for (k=1;k<=ld1;k++)
           WP[1].addterm(cf[k],k+1);
        for (v=2;v<=ld;v++)
            WP[v]=modxn(WP[v-1]*WP[1],ld+1);
//
// WPv have understood multiplier x^-v
//            
        get_ck(ld1,atilde,btilde,cft);

        Y=0;
        for (k=1;k<=ld1;k++)
            Y.addterm((lp*cf[k]-cft[k])/(ZZn)((2*k+1)*(2*k+2)),k+1);
        Y.addterm(-p1,1);

        RF=1;
        H=1;
        X=1;
        for (r=1;r<=ld;r++)
        {
            X=modxn(X*Y,ld+1);
            RF*=r;
            H+=(X/RF);
        }
//
//  H has understood multiplier x^-d
//
        ad=1;
        Fl=0;
        Fl.addterm(ad,ld);
        for (v=ld-1;v>=0;v--)
        {
            H-=ad*WP[v+1];
            H=divxn(H,1);
            ad=H.min();
            Fl.addterm(ad,v);
        }

        setmod(Fl);
        MY2=Y2;
        MY4=Y4;

        XP=pow(XX,p);
        YP=pow(MY2,(p-1)/2);

//
// Now looking for value of lambda which satisfies
// (X^P,Y^P) = lambda.(XX,YY)
//
        good[nl]=lp;
        cout << "NP mod " << lp << " = " << flush;
        XT=XP;
        YT=YP;
        ZT=one;
        for (lambda=1;lambda<lp;lambda++)
        {
            tau=(lambda+invers(lambda,lp)*p)%lp;
            cout << setw(3) << (p+1-tau)%lp << flush; 
            elliptic_add(XT,YT,ZT,XX,YY);
            if (iszero(ZT)) 
            {  // Is it point at infinity?
                t[nl]=tau;
                if ((p+1-tau)%lp==0)
                {
                    cout << " ***" << endl;
                    if (search) escape=TRUE;
                }
                else cout << endl;
                break;
            }
            cout << "\b\b\b";
        }
        nl++;        
        accum*=lp;
        if (escape) break;
    }

    if (escape) {b+=1; continue;}

    Crt CRT(nl,good);
    Big order,ordermod;
    ordermod=accum;
    order=(p+1-CRT.eval(t))%ordermod;    // get order mod product of primes

    nrp=kangaroo(p,order,ordermod);

    if (!prime(nrp) && search) {b+=1; continue; }
    else break;
    }

    if (fout) 
    {
        ECn P;
        ofile << a << endl;
        ofile << b << endl;
    // generate a random point on the curve 
        do {
            x=rand(p);
        } while (!P.set(x,x));
        P.get(x,y);
        ofile << x << endl;
        ofile << y << endl;
        ofile << p << endl;
        ofile << nrp << endl;
    }
    return 0;
}

